import logging
from typing import Any, Callable, List, Optional, Union

import numpy as np
import pandas as pd
from sam.logging_functions import log_dataframe_characteristics, log_new_columns
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

logger = logging.getLogger(__name__)


def _nfft_helper(
    series: pd.Series,
    nfft: Callable[[np.ndarray, np.ndarray], np.ndarray],
) -> np.ndarray:
    """Helper function to apply nfft to series.

    Parameters
    ----------
    series : pd.Series
        series to apply nfft on.
    nfft : function
        nfft function.

    Returns
    -------
    np.ndarray
        result after applying nfft on series.
    """
    # series should be even sample, cutoff one if it is not
    if series.size % 2 != 0:
        series = series.iloc[1:]
    if series.size <= 5:  # too short
        return np.array([])

    # first convert time index to nanoseconds since epoch
    # If a non-pandas timestamp index is used, these units aren't nanoseconds
    # However, since we normalize to [-0.5, 0.5] anyway, it doesn't matter
    time = np.array(series.index.astype(int))
    # then make first value 0
    time -= np.min(time)
    # normalize to run from -0.5 to 0.5
    time = time / np.max(time) - 0.5

    # now do fft on normalized values
    f = nfft(time, series.values - np.mean(series.values))
    # flip fft spectrum so it matches the numpy spectrum
    ff = np.fft.fftshift(f)
    # Only take first half since that is useful. See documentation inside
    # multicol_output.Helper.__init__ for better explanation
    return np.abs(ff)[1 : len(ff) // 2]


def multicol_output(
    arr: np.ndarray,
    n: int,
    func: Callable[[np.ndarray], np.ndarray],
    fourier: bool = False,
    time_window: str = None,
) -> pd.DataFrame:
    """
    Generic function to compute multiple columns
    func is a function that takes in a numpy array, and outputs a numpy array of the same length
    The numpy array will be a window: e.g. if n=3, func will get a window of 3 points, and output
    3 values. Then, those 3 values will be converted to columns in a dataframe.
    For fourier, an additional column selection is done, so less than 3 columns would be returned
    In that case, the option 'fourier' needs to be set to True

    For nfft, we need not only a numpy array of values, but also a DatetimeIndex. Additionally, we
    have a time_window instead of an integer window. This time_window is for example '100min'.
    If time_window is set, this function will assume that arr has a DatetimeIndex. n is still used
    as the output size.

    Parameters
    ----------
    arr : np.ndarray
        input numpy array to generate multipole columns from.
    n : int
        window size in number of datapoints.
    func : function
        aggregate function to apply to the rolling series.
    fourier : bool, optional
        enable fourier transformation, by default False.
    time_window : str, optional
        size of the time window in pandas timedelta format, by default None.

    Returns
    -------
    pd.DataFrame
        Result after applying a rolling function with size `n` and aggregate function `func`.
    """

    class Helper:
        # https://stackoverflow.com/a/39064656
        def __init__(self, nrow, n):
            if fourier:
                """we are only interested in these coefficients, since the rest is redundant
                See https://en.wikipedia.org/wiki/Discrete_Fourier_transform
                It follows from the definition that when k = 0, the result is simply the sum
                It also follows that the definition for n-k is the same as k, except inverted
                Since we take the absolute value, this inversion is removed, so the result is
                identical. Therefore, we only want the values from k = 1 to k = n//2
                """
                self.useful_coeffs = range(1, n // 2 + 1)
            else:
                self.useful_coeffs = range(0, n)
            ncol = len(self.useful_coeffs)
            self.series = np.full((nrow, ncol), np.nan)
            if time_window is None:
                self.calls = n - 1
            else:
                self.calls = 0
            self.n = n

        def calc_func(self, vector):
            if len(vector) < self.n and time_window is None:
                # We are still at the beginning of the dataframe, nothing to do
                return np.nan
            values = func(vector)
            # If function did not return enough values, pad with zeros
            values = np.pad(values, (0, max(0, self.n - values.size)), "constant")
            values = values[self.useful_coeffs]
            self.series[self.calls, :] = values
            self.calls = self.calls + 1
            return np.nan  # return something to make Rolling apply not error

    helper = Helper(len(arr), n)
    if time_window is None:
        arr.rolling(n, min_periods=0).apply(helper.calc_func, raw=True)
    else:
        # For time-window, we need raw=False because 'func' may need
        # The DatetimeIndex. Even though raw=False is slower.
        arr.rolling(time_window).apply(helper.calc_func, raw=False)
    return pd.DataFrame(helper.series)


class BuildRollingFeatures(BaseEstimator, TransformerMixin):
    """Applies some rolling function to a pandas dataframe

    This class provides a stateless transformer that applies to each column in a dataframe.
    It works by applying a certain rolling function to each column individually, with a
    window size. The rolling function is given by rolling_type, for example 'mean',
    'median', 'sum', etcetera.

    An important note is that this transformer assumes that the data is sorted by time already!
    So if the input dataframe is not sorted by time (in ascending order), the results will be
    completely wrong.

    A note about the way the output is rolled: in case of 'lag' and 'diff', the output will
    always be lagged, even if lookback is 0. This is because these functions inherently look
    at a previous cell, regardless of what the lookback is. All other functions will start
    by looking at the current cell if lookback is 0. (and will also look at previous cells
    if `window_size` is greater than 1)

    'ewm' looks at `window_size` a bit different: instead of a discrete number of points to
    look at, 'ewm' needs a parameter alpha between 0 and 1 instead.

    Parameters
    ----------
    window_size: array-like, shape = (n_outputs, ), optional (default=None)
        vector of values to shift. Ignored when rolling_type is ewm
        if integer, the window size is fixed, and the timestamps are assumed to be uniform.
        If string of timeoffset (for example '1H'), the input dataframe must have a DatetimeIndex.
        timeoffset is not supported for rolling_type 'lag', 'fourier', 'ewm', 'diff'!
    lookback: number type, optional (default=1)
        the features that are built will be shifted by this value
        If more than 0, this prevents leakage
    rolling_type: string, optional (default="mean")
        The rolling function. Must be one of: 'median', 'skew', 'kurt', 'max', 'std', 'lag',
        'mean', 'diff', 'sum', 'var', 'min', 'numpos', 'ewm', 'fourier', 'cwt', 'trimmean'
    deviation: str, optional (default=None)
        one of ['subtract', 'divide']. If this option is set, the resulting column will either
        have the original column subtracted, or will be divided by the original column. If None,
        just return the resulting column. This option is not allowed when rolling_type is 'cwt'
        or 'fourier', but it is allowed with all other rolling_types.
    alpha: numeric, optional (default=0.5)
        if rolling_type is 'ewm', this is the parameter alpha used for weighing the samples.
        The current sample weighs alpha, the previous sample weighs alpha*(1-alpha), the
        sample before that weighs alpha*(1-alpha)^2, etcetera. Must be in (0, 1]
    width: numeric, optional (default=1)
        if rolling_type is 'cwt', the wavelet transform uses a ricker signal. This parameter
        defines the width of that signal
    nfft_ncol: numeric, optional (default=10)
        if rolling_type is 'nfft', there needs to be a fixed number of columns as output, since
        this is unknown a-priori. This means the number of output-columns will be fixed. If
        nfft has more outputs, and additional outputs are discarded. If nfft has less outputs,
        the rest of the columns are right-padded with 0.
    proportiontocut: numeric, optional (default=0.1)
        if rolling_type is 'trimmean', this is the parameter used to trim values on both tails
        of the distribution. Must be in [0, 0.5). Value 0 results in the mean, close to 0.5
        approaches the median.
    keep_original: boolean, optional (default=True)
        if the original columns should be kept or discarded
        True by default, which means the new columns are added to the old ones
    timecol: str, optional (default=None)
        Optional, the column to set as the index during transform. The index is restored before
        returning. This is only useful when using a timeoffset for window_size, since that needs
        a datetimeindex. So this column can specify a time column. This column will not be
        feature-engineered, and will never be returned in the output!
    add_lookback_to_colname: bool, optional (default=False)
        Whether to add lookback to the newly generated column names.
        if False, column names will be like: DEBIET#mean_2
        if True, column names will be like: DEBIET#mean_2_lookback_0

    Examples
    --------
    >>> from sam.feature_engineering import BuildRollingFeatures
    >>> import pandas as pd
    >>> df = pd.DataFrame({'RAIN': [0.1, 0.2, 0.0, 0.6, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    ...                    'DEBIET': [1, 2, 3, 4, 5, 5, 4, 3, 2, 4, 2, 3]})
    >>>
    >>> BuildRollingFeatures(rolling_type='lag', window_size = [0,1,4], \\
    ...                      lookback=0, keep_original=False).fit_transform(df)
        RAIN#lag_0  DEBIET#lag_0  ...  RAIN#lag_4  DEBIET#lag_4
    0          0.1             1  ...         NaN           NaN
    1          0.2             2  ...         NaN           NaN
    2          0.0             3  ...         NaN           NaN
    3          0.6             4  ...         NaN           NaN
    4          0.1             5  ...         0.1           1.0
    5          0.0             5  ...         0.2           2.0
    6          0.0             4  ...         0.0           3.0
    7          0.0             3  ...         0.6           4.0
    8          0.0             2  ...         0.1           5.0
    9          0.0             4  ...         0.0           5.0
    10         0.0             2  ...         0.0           4.0
    11         0.0             3  ...         0.0           3.0
    <BLANKLINE>
    [12 rows x 6 columns]
    """

    def __init__(
        self,
        rolling_type: str = "mean",
        lookback: int = 1,
        window_size: Optional[str] = None,
        deviation: Optional[str] = None,
        alpha: float = 0.5,
        width: int = 1,
        nfft_ncol: int = 10,
        proportiontocut: float = 0.1,
        timecol: Optional[str] = None,
        keep_original: bool = True,
        add_lookback_to_colname: bool = False,
    ):

        self.window_size = window_size
        self.lookback = lookback
        self.rolling_type = rolling_type
        self.deviation = deviation
        self.alpha = alpha
        self.width = width
        self.nfft_ncol = nfft_ncol
        self.proportiontocut = proportiontocut
        self.keep_original = keep_original
        self.timecol = timecol
        self.add_lookback_to_colname = add_lookback_to_colname
        logger.debug(
            "Initialized rolling generator. rolling_type={}, lookback={}, "
            "window_size={}, deviation={}, alpha={}, proportiontocut={}, width={}, "
            "keep_original={}, timecol={}".format(
                rolling_type,
                lookback,
                window_size,
                deviation,
                alpha,
                proportiontocut,
                width,
                keep_original,
                timecol,
            )
        )

    def _validate_params(self):
        """apply various checks to the inputs of the __init__ function
        throw value error or type error based on the result
        """

        self._validate_lookback()
        self._validate_width()
        self._validate_alpha()
        self._validate_proportiontocut()

        if not isinstance(self.rolling_type, str):
            raise TypeError("rolling_type must be a string")

        if not isinstance(self.keep_original, bool):
            raise TypeError("keep_original must be a boolean")

        if int(self.nfft_ncol) != self.nfft_ncol:
            raise ValueError("nfft_ncol must be an integer!")

        if self.deviation not in [None, "subtract", "divide"]:
            raise ValueError("Deviation must be one of [None, 'subtract', 'divide']")

        if self.window_size is None and self.rolling_type != "ewm":
            raise ValueError("Window_size must not be None, unless rolling_type is ewm")

        if self.deviation is not None and self.rolling_type in ["fourier", "cwt"]:
            raise ValueError("Deviation cannot be used together with {}".format(self.rolling_type))

    def _validate_lookback(self):
        if not np.isscalar(self.lookback):
            raise TypeError("lookback must be a scalar")
        if self.lookback < 0:
            raise ValueError("lookback cannot be negative!")

    def _validate_width(self):
        if not np.isscalar(self.width):
            raise TypeError("width must be a scalar")
        if self.width <= 0:
            raise ValueError("width must be positive")

    def _validate_alpha(self):
        if not np.isscalar(self.alpha):
            raise TypeError("alpha must be a scalar")
        if self.alpha <= 0 or self.alpha > 1:
            raise ValueError("alpha must be in (0, 1]")

    def _validate_proportiontocut(self):
        if not np.isscalar(self.proportiontocut):
            raise TypeError("proportiontocut must be a scalar")
        if self.proportiontocut >= 0.5 or self.proportiontocut < 0:
            raise ValueError("proportiontocut must be in [0, 0.5)")

    def _get_rolling_fun(
        self,
        rolling_type: str = "mean",
    ) -> Callable[[Union[pd.Series, np.ndarray], Union[int, None]], Union[pd.Series, np.ndarray]]:
        """Given a function name as a string, creates a function that
        applies that rolling function

        Parameters
        ----------
        rolling_type : string, default="mean"
            the description of the rolling function. must be one of lag, sum, mean,
            median, trimmean, var, std, max, min, skew, kurt, diff, numpos, ewm, fourier, cwt

        Returns
        -------
        rolling_function : function
            function with two inputs: a pandas series and an integer. Will apply
            some rolling function to the series, with window size of the integer.
            Alternatively, in case of fourier/cwt, a function with one input:
            a numpy array. Will output another numpy array.
        """
        if self.rolling_type == "cwt":
            from scipy import signal  # Only needed for this rolling type
        if self.rolling_type == "nfft":
            from nfft import nfft
        if self.rolling_type == "trimmean":
            from scipy.stats import trim_mean

        rolling_functions = {
            "lag": lambda arr, n: arr.shift(n),
            "sum": lambda arr, n: arr.rolling(n).sum(),
            "mean": lambda arr, n: arr.rolling(n).mean(),
            "trimmean": lambda arr, n: arr.rolling(n).apply(
                lambda w: trim_mean(w, self.proportiontocut), raw=True
            ),
            "median": lambda arr, n: arr.rolling(n).median(),
            "var": lambda arr, n: arr.rolling(n).var(),
            "std": lambda arr, n: arr.rolling(n).std(),
            "max": lambda arr, n: arr.rolling(n).max(),
            "min": lambda arr, n: arr.rolling(n).min(),
            "skew": lambda arr, n: arr.rolling(n).skew(),
            "kurt": lambda arr, n: arr.rolling(n).kurt(),
            "diff": lambda arr, n: arr.diff(n),
            "numpos": lambda arr, n: arr.gt(0).rolling(n).sum(),
            "ewm": lambda arr, n: arr.ewm(alpha=self.alpha).mean(),
            # These two have different signature because they are called by multicol_output
            "fourier": lambda vector: np.absolute(np.fft.fft(vector)),
            "cwt": lambda vector: signal.cwt(vector, signal.ricker, [self.width])[0],
            "nfft": lambda series: _nfft_helper(series, nfft),
        }

        if rolling_type not in rolling_functions:
            raise ValueError(
                "The rolling_type is %s, which is not an available function" % rolling_type
            )

        return rolling_functions[rolling_type]

    def _apply_deviation(
        self, arr: np.ndarray, original: np.ndarray, deviation: str
    ) -> np.ndarray:
        """Helper function to apply deviation during the transform"""
        if deviation is None:
            return arr
        if deviation == "subtract":
            return arr - original
        if deviation == "divide":
            # Pandas will insert inf when dividing by 0
            return arr / original

    def _generate_and_add_new_features(
        self, X: pd.DataFrame, result: pd.DataFrame
    ) -> pd.DataFrame:
        """Applies rolling functions to pandas dataframe `X` and concatenates result to `result`.

        Parameters:
        ----------
        X: pandas dataframe
           the pandas dataframe that you want to apply rolling functions on
        result: pandas dataframe
           the pandas dataframe that you want to add the new features to

        Returns
        -------
        pandas dataframe, shape = `(n_rows, n_features * (n_outputs + 1))`
            the pandas dataframe, appended with the new feature columns
        """

        if self.rolling_type in ["fourier", "cwt", "nfft"]:
            for window_size, suffix in zip(self.window_size_, self.suffix_):
                # If rolling type is nfft, the time_window needs to be set
                if self.rolling_type == "nfft":
                    window_size, time_window = self.nfft_ncol, window_size
                else:
                    time_window = None

                for column in X.columns:
                    new_features = multicol_output(
                        X[column],
                        window_size,
                        self.rolling_fun_,
                        self.rolling_type == "fourier",
                        time_window=time_window,
                    ).shift(self.lookback)
                    # Fourier has less columns
                    if self.rolling_type == "fourier":
                        useful_coeffs = range(1, window_size // 2 + 1)
                    else:
                        useful_coeffs = range(0, window_size)
                    col_prefix = "#".join([str(column), suffix])
                    new_features.columns = ["_".join([col_prefix, str(j)]) for j in useful_coeffs]
                    new_features = new_features.set_index(X.index)
                    result = pd.concat([result, new_features], axis=1)
        else:
            for window_size, suffix in zip(self.window_size_, self.suffix_):
                new_features = X.apply(
                    lambda arr: self._apply_deviation(
                        self.rolling_fun_(arr, window_size).shift(self.lookback),
                        arr,
                        self.deviation,
                    ),
                    raw=False,
                )
                new_features.columns = [
                    "#".join([str(col), suffix]) for col in new_features.columns
                ]
                result = pd.concat([result, new_features], axis=1)

        return result

    def fit(self, X: Any = None, y: Any = None):
        """Calculates window_size and feature function

        Parameters
        ----------
        X: optional, is ignored
        y: optional, is ignored
        """

        self._validate_params()

        if self.rolling_type == "ewm":
            # ewm needs no integer window_size
            self.window_size_ = "ewm"
            self.suffix_ = ["ewm_" + str(self.alpha)]
        else:
            self.window_size_ = self.window_size
            # Singleton window_size is also allowed
            if np.isscalar(self.window_size_):
                self.window_size_ = [self.window_size_]
            self.suffix_ = [
                self.rolling_type + "_" + str(window_size) for window_size in self.window_size_
            ]
            if self.add_lookback_to_colname:
                self.suffix_ = [s + "_lookback_" + str(self.lookback) for s in self.suffix_]
        self.rolling_fun_ = self._get_rolling_fun(self.rolling_type)
        logger.debug(
            "Done fitting transformer. window size: {}, suffix: {}".format(
                self.window_size_, self.suffix_
            )
        )

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transforms pandas dataframe `X` to apply rolling function

        Parameters
        ----------
        X: pandas dataframe, shape = `(n_rows, n_features)`
           the pandas dataframe that you want to apply rolling functions on

        Returns
        -------
        result: pandas dataframe, shape = `(n_rows, n_features * (n_outputs + 1))`
            the pandas dataframe, appended with the new columns
        """

        check_is_fitted(self, "window_size_")
        if self.keep_original:
            result = X.copy()
        else:
            result = pd.DataFrame(index=X.index)

        if self.timecol is not None:
            # Set DatetimeIndex on the intermediate result
            index_backup = X.index.copy()
            new_index = pd.DatetimeIndex(X[self.timecol].values)
            X = X.set_index(new_index).drop(self.timecol, axis=1)
            if self.keep_original:
                result = result.set_index(new_index).drop(self.timecol, axis=1)
            else:
                result = pd.DataFrame(index=new_index)

        result = self._generate_and_add_new_features(X, result)

        self._feature_names = list(result.columns.values)
        if self.timecol is not None:
            result = result.set_index(index_backup)
        log_new_columns(result, X)
        log_dataframe_characteristics(result, logging.DEBUG)  # Log types as well

        return result

    def get_feature_names_out(self, input_features=None) -> List[str]:
        """
        Returns feature names for the outcome of the last transform call.
        """
        check_is_fitted(self, "_feature_names")

        return self._feature_names



import numpy as np
from sam.metrics import precision_incident_recall_curve
from sklearn.metrics import precision_recall_curve


def plot_threshold_curve(y_true: np.array, y_score: np.array, range_pred: tuple = None):
    """
    Create and return a threshold curve, also known as Ynformed plot. It does this by putting
    the threshold on the x-axis, and for each threshold, plotting the precision and recall.
    This returns a subplot object that can be shown or edited further.

    Parameters
    ----------
    y_true: array_like, shape = (n_outputs,)
        The actual values. Values must be either 0 or 1. If range_pred is provided, this refers
        to the incidents.
    y_score: array_like, shape = (n_outputs,)
        The prediction. Values must be between 0 and 1
    range_pred: tuple, optional (default = None)
        If provided, make a precision/incident recall plot, using y_true as the
        incidents, and this range_pred.

    Returns
    -------
    plot:  matplotlib.axes._subplots.AxesSubplot object
        a plot containing the resulting precision threshold plot.
        can be edited further, or printed to the output.

    Examples
    --------
    >>> from sam.visualization import plot_threshold_curve
    >>> plot_threshold_curve([0, 1, 0, 1, 1, 0], [.2, .3, .4, .5, .9, .1])  # doctest: +SKIP

    >>> # Incident recall threshold plot
    >>> plot_threshold_curve(
    ... [0, 0, 0, 0, 1, 0], [.2, .3, .4, .5, .9, .1], (0, 1)
    ... )  # doctest: +SKIP
    """
    import matplotlib.pyplot as plt

    if range_pred is None:
        # Retrieve the curve data
        precision, recall, thresholds = precision_recall_curve(y_true, y_score)
        recall_label = "Recall"
    else:
        precision, recall, thresholds = precision_incident_recall_curve(
            y_true, y_score, range_pred
        )
        recall_label = "Incident Recall"

    # Initialize the figure
    _, ax = plt.subplots()
    ax.set_title("Precision and Recall Scores per threshold")

    # Plot the curves in the figure
    ax.plot(thresholds, precision[:-1], label="Precision")
    ax.plot(thresholds, recall[:-1], label=recall_label)
    ax.set_ylabel("Score")
    ax.set_xlabel("Decision Threshold")
    ax.legend(loc="best")

    return ax



import numpy as np
from sam.metrics import precision_incident_recall_curve
from sklearn.metrics import average_precision_score, precision_recall_curve


def plot_precision_recall_curve(
    y_true: np.array,
    y_score: np.array,
    range_pred: tuple = None,
    color: str = "b",
    alpha: float = 0.2,
):
    """
    Create and return a precision-recall curve. It does this by putting the
    precision on the y-axis, the recall on the x-axis, and for each threshold,
    plotting that precision-recall.
    The function returns a subplot object that can be shown or edited further.

    Parameters
    ----------
    y_true : np.array, shape = (n_outputs,)
        The truth values. Must be either 0 or 1. If range_pred is provided, this refers
        to the incidents.
    y_score: np.array, shape = (n_outputs,)
        The prediction. Must be between 0 and 1
    range_pred : tuple, optional (default = None)
        If this is provided, make a precision/incident recall plot, using y_true as the
        incidents, and this range_pred.
    color: string, optional (default='b')
        The color the the plot. Default is blue
    alpha: float, optional (default=0.2)
        The transparency of the solid part of the plot.
        The line will always be alpha 0.1, so if you
        just want the line, set this to 0.

    Returns
    -------
    plot:  matplotlib.axes._subplots.AxesSubplot object
        a plot containing the resulting precision recall_curve.
        can be edited further, or printed to the output.

    Examples
    --------
    >>> from sam.visualization import plot_precision_recall_curve
    >>> y_true = np.array([0, 0, 1, 1, 1, 0])
    >>> y_scores = np.array([0.1, 0.4, 0.35, 0.8, 0.2, 0.3])
    >>>
    >>> fig = plot_precision_recall_curve(y_true, y_scores)

    >>> # Incident recall curve
    >>> y_incidents = np.array([0, 0, 0, 0, 1, 0])
    >>> fig2 = plot_precision_recall_curve(y_incidents, y_scores, (0, 2))
    """
    import matplotlib.pyplot as plt

    if range_pred is None:
        precision, recall, _ = precision_recall_curve(y_true, y_score)
        ap_score = average_precision_score(y_true, y_score)
        title = "2-class Precision-Recall curve. AP={0:0.2f}".format(ap_score)
        recall_label = "Recall"
    else:
        precision, recall, _ = precision_incident_recall_curve(y_true, y_score, range_pred)
        title = "2-class Precision-Incident Recall curve. Prediction range: ({}, {})".format(
            range_pred[0], range_pred[1]
        )
        recall_label = "Incident Recall"

    plt.figure()
    ax = plt.subplot()
    ax.step(recall, precision, color=color, alpha=1, where="post")
    ax.fill_between(recall, precision, alpha=alpha, color=color, step="post")

    ax.set_xlabel(recall_label)
    ax.set_ylabel("Precision")
    ax.set_ylim([0.0, 1.05])
    ax.set_xlim([0.0, 1.0])
    ax.set_title(title)
    return ax



import logging
from typing import Callable, Union

import numpy as np
import pandas as pd
from sam.feature_engineering.rolling_features import BuildRollingFeatures
from sam.logging_functions import log_dataframe_characteristics

logger = logging.getLogger(__name__)


def lag_correlation(
    df: pd.DataFrame,
    target_name: str,
    lag: int = 12,
    method: Union[str, Callable] = "pearson",
):
    """
    Creates a new dataframe that contains the correlation of target_name
    with other variables in the dataframe based on the output from
    BuildRollingFeatures. The results are processed for easy visualization,
    with a column for the lag and then correlation per feature.

    Parameters
    ----------
    df: pd.DataFrame
        input dataframe contains variables to calculate lag correlation of
    target_name: str
        The name of the goal variable to calculate lag correlation with
    lag: int or list of ints (default=12)
        When an integer is provided, a range is created from 0 to lag in steps
        of 1, when an array of ints is provided, this is directly used.
        Default is 12, which means the correlation is calculated for lag
        ranging from 0 to 11.
    method: string or callable, optional (default='pearson')
        The method used to calculate correlation. See pandas.DataFrame.corrwith.
        Options are {‘pearson’, ‘kendall’, ‘spearman’}, or a callable.

    Returns
    -------
    tab: pandas dataframe
        A dataframe with the correlations and shift.
        The column header contains the feature name.

    Examples
    --------
    >>> import pandas as pd
    >>> from sam.exploration import lag_correlation
    >>> import numpy as np
    >>> X = pd.DataFrame({
    ...        'RAIN': [0.1, 0.2, 0.0, 0.6, 0.1, 0.0, 0.0,
    ...                 0.0, 0.0, 0.0, 0.0, 0.0],
    ...        'DEBIET#A': [1, 2, 3, 4, 5, 5, 4, 3, 2, 4, 2, 3],
    ...        'DEBIET#B': [3, 1, 2, 3, 3, 6, 4, 1, 3, 3, 1, 5]})
    >>> X['DEBIET#TOTAAL'] = X['DEBIET#A'] + X['DEBIET#B']
    >>> tab = lag_correlation(X, 'DEBIET#TOTAAL', lag=11)
    >>> tab
        LAG  DEBIET#A  DEBIET#B      RAIN
    0     0  0.838591  0.897340 -0.017557
    1     1  0.436484  0.102808  0.204983
    2     2  0.287863 -0.401768  0.672316
    3     3 -0.388095 -0.140876  0.188438
    4     4 -0.632980 -0.509307 -0.227071
    5     5 -0.667537 -0.367268 -0.048162
    6     6 -0.152832  0.615239  0.110876
    7     7  0.457496 -0.107833 -0.719702
    8     8  0.291111  0.039253  0.871695
    9     9  0.188982  0.755929 -0.944911
    10   10  1.000000 -1.000000  1.000000
    """

    logging.debug("Now creating lag correlation with lag {}".format(lag))
    y = df[target_name]
    X = df.drop(target_name, axis=1)

    if np.isscalar(lag):
        lag = np.arange(lag)

    RollingFeatures = BuildRollingFeatures(
        rolling_type="lag", window_size=lag, lookback=0, keep_original=False
    )
    df = RollingFeatures.fit_transform(X)
    corr_table = df.corrwith(y, method=method).reset_index()
    corr_table.columns = ["index", "corr"]
    corr_table["LAG"] = corr_table["index"].apply(lambda x: x.rsplit("_", 1)[1])
    corr_table["GROUP"] = corr_table["index"].apply(lambda x: x.rsplit("#", 1)[0])

    tab = pd.pivot_table(corr_table, values="corr", index="LAG", columns="GROUP").reset_index()
    tab["LAG"] = pd.to_numeric(tab["LAG"])
    tab.columns.name = None
    tab = tab.sort_values("LAG").reset_index(drop=True)

    logging.info("create_lag_correlation output:")
    log_dataframe_characteristics(tab, logging.INFO)
    return tab



import pandas as pd
from sam.validation import MADValidator


def diagnostic_extreme_removal(
    mad_validator: MADValidator,
    raw_data: pd.DataFrame,
    col: str,
):
    """
    Creates a diagnostic plot for the extreme value removal procedure.

    Parameters:
    ----------
    mad_validator: sam.validation.MADValidator
        fitted MADValidator object
    raw_data: pd.DataFrame
        non-transformed data data
    col: string
        column name to plot

    Returns:
    -------
    fig: matplotlib.pyplot.figure
        diagnostic plot
    """
    import matplotlib.pyplot as plt
    import seaborn as sns

    # get data
    x = raw_data[col].copy()
    invalid_w = mad_validator.validate(raw_data)[col]
    invalid_values = x.loc[invalid_w]
    rolling = mad_validator._compute_rolling(x)
    diff = x.values - rolling

    # generate plot
    fig = plt.figure(figsize=(12, 6))
    plt.subplot(211)
    plt.title(col)

    plt.plot(x.index, x.values, label="original_signal", lw=5)

    plt.plot(
        rolling.index,
        rolling.values,
        "--k",
        label="rolling median",
        lw=3,
    )

    plt.plot(
        invalid_values.index,
        invalid_values.values,
        "o",
        ms=10,
        mew=2,
        mec="r",
        fillstyle="none",
        label="invalid samples",
    )

    plt.legend(loc="best")
    sns.despine()

    plt.subplot(212)
    plt.plot(diff.values, label="abs(original - rolling)")
    plt.axhline(mad_validator.thresh_high[col], ls="--", c="r")
    plt.axhline(mad_validator.thresh_low[col], ls="--", c="r", label="thresholds")
    plt.legend(loc="best")
    sns.despine()
    plt.tight_layout()

    return fig



import pandas as pd
from sam.validation import FlatlineValidator


def diagnostic_flatline_removal(
    flatline_validator: FlatlineValidator,
    raw_data: pd.DataFrame,
    col: str,
):
    """
    Creates a diagnostic plot for the extreme value removal procedure.

    Parameters:
    ----------
    flatline_validator: sam.validation.FlatlineValidator
        fitted FlatlineValidator object
    raw_data: pd.DataFrame
        non-transformed data
    col: string
        column name to plot

    Returns:
    -------
    fig: matplotlib.pyplot.figure
        diagnostic plot
    """
    import matplotlib.pyplot as plt

    # get data
    x = raw_data[col].copy()
    invalid_w = flatline_validator.validate(raw_data)[col]
    invalid_values = x[invalid_w]

    # generate plot
    fig = plt.figure(figsize=(12, 6))
    plt.subplot(111)
    plt.title(col)

    plt.plot(x.index, x.values, label="original_signal", lw=5)

    plt.plot(
        invalid_values.index,
        invalid_values.values,
        "o",
        ms=10,
        mew=2,
        mec="r",
        fillstyle="none",
        label="invalid samples",
    )

    plt.legend(loc="best")
    plt.tight_layout()

    return fig



from typing import Union

import numpy as np
import pandas as pd
from sklearn.base import clone
from sklearn.preprocessing import StandardScaler


def scale_train_test(
    X_train: pd.DataFrame,
    X_test: pd.DataFrame,
    y_train: Union[pd.DataFrame, pd.Series],
    y_test: Union[pd.DataFrame, pd.Series],
    scaler=StandardScaler(),
):
    """
    Fit a transformer to the train sets to scale and later invert scaling
    of the data to a distribution on which the algorithm can learn faster.

    Parameters:
    ----------
    X_train: pd.DataFrame
        containing the training features
    X_test: pd.DataFrame
        containing the test features
    y_train: pd.Series or pd.DataFrame
        containing the train target
    y_test: pd.Series or pd.DataFrame
        containing the test target
    scaler: sklearn scaler instance
        e.g. StandardScaler(),  MinMaxScaler(), or RobustScaler()
        for available options see:
        <https://scikit-learn.org/stable/auto_examples/
        preprocessing/plot_all_scaling.html>

    Returns:
    -------
    X_train: pd.DataFrame
        rescaled dataframe containing the train feature samples
    X_test: pd.DataFrame
        rescaled dataframe containing the test feature samples
    y_train: pd.Series or pd.DataFrame
        rescaled series containing the train target samples
    y_train: pd.Series or pd.DataFrame
        rescaled series containing the test target samples
    X_scaler: sklearn scaler instance
        fitted X scaler instance that can be used to invert scaling
    y_scaler: sklearn scaler instance
        fitted y scaler instance that can be used to invert scaling
    """

    X_scaler, y_scaler = clone(scaler), clone(scaler)

    X_train = pd.DataFrame(
        X_scaler.fit_transform(X_train), columns=X_train.columns, index=X_train.index
    )
    X_test = pd.DataFrame(X_scaler.transform(X_test), columns=X_test.columns, index=X_test.index)

    # we only need to test if train set is series of dataframe,
    # as this needs to be the same for the test set
    if isinstance(y_train, pd.Series):
        y_train = pd.Series(
            np.ravel(y_scaler.fit_transform(np.array(y_train).reshape(-1, 1))),
            index=y_train.index,
        )
        y_test = pd.Series(
            np.ravel(y_scaler.transform(np.array(y_test).reshape(-1, 1))),
            index=y_test.index,
        )
    elif isinstance(y_train, pd.DataFrame):
        y_train = pd.DataFrame(
            y_scaler.fit_transform(y_train),
            columns=y_train.columns,
            index=y_train.index,
        )
        y_test = pd.DataFrame(
            y_scaler.transform(y_test), columns=y_test.columns, index=y_test.index
        )

    return X_train, X_test, y_train, y_test, X_scaler, y_scaler



from pathlib import Path
from typing import Callable, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from sam.feature_engineering import BaseFeatureEngineer
from sam.metrics import R2Evaluation, keras_joint_mse_tilted_loss
from sam.models import create_keras_quantile_mlp
from sam.models.base_model import BaseTimeseriesRegressor
from sam.preprocessing import make_shifted_target
from sklearn.base import TransformerMixin
from sklearn.utils.validation import check_is_fitted
from sam.models.sam_shap_explainer import SamShapExplainer


class MLPTimeseriesRegressor(BaseTimeseriesRegressor):
    """Multi-layer Perceptron Regressor for time series

    This model combines several approaches to time series data:
    Multiple outputs for forecasting, quantile regression, and feature engineering.
    This class is an implementation of an MLP to estimate multiple quantiles for all
    forecasting horizons at once.

    This is a wrapper for a keras MLP model. For more information on the model parameters,
    see the keras documentation.

    Parameters
    ----------
    predict_ahead: tuple of integers, optional (default=(0,))
        how many steps to predict ahead. For example, if (1, 2), the model will predict both 1 and
        2 timesteps into the future. If (0,), predict the present.
    quantiles: tuple of floats, optional (default=())
        The quantiles to predict. Values between 0 and 1. Keep in mind that the mean will be
        predicted regardless of this parameter
    use_diff_of_y: bool, optional (default=True)
        If True differencing is used (the difference between y now and shifted y),
        else differencing is not used (shifted y is used).
    timecol: string, optional (default=None)
        If not None, the column to use for constructing time features. For now,
        creating features from a DateTimeIndex is not supported yet.
    y_scaler: object, optional (default=None)
        Should be an sklearn-type transformer that has a transform and inverse_transform method.
        E.g.: StandardScaler() or PowerTransformer()
    feature_engineering: object, optional (default=None)
        Should be an sklearn-type transformer that has a transform method, e.g.
        `sam.feature_engineering.SimpleFeatureEngineer`.
    n_neurons: integer, optional (default=200)
        The number of neurons to use in the model, see :ref:`create_keras_quantile_mlp
        <create-keras-quantile-mlp>`
    n_layers: integer, optional (default=2)
        The number of layers to use in the model, see :ref:`create_keras_quantile_mlp
        <create-keras-quantile-mlp>`
    batch_size: integer, optional (default=16)
        The batch size to use in the model, see :ref:`create_keras_quantile_mlp
        <create-keras-quantile-mlp>`
    epochs: integer, optional (default=20)
        The number of epochs to use in the model, see :ref:`create_keras_quantile_mlp
        <create-keras-quantile-mlp>`
    lr: float, optional (default=0.001)
        The learning rate to use in the model, see :ref:`create_keras_quantile_mlp
        <create-keras-quantile-mlp>`
    dropout: float, optional (default=None)
        The type of dropout to use in the model, see :ref:`create_keras_quantile_mlp
        <create-keras-quantile-mlp>`
    momentum: integer, optional (default=None)
        The type of momentum in the model, see :ref:`create_keras_quantile_mlp
        <create-keras-quantile-mlp>`
    verbose: integer, optional (default=1)
        The verbosity of fitting the keras model. Can be either 0, 1 or 2.
    r2_callback_report: boolean (default=False)
        Whether to add train and validation r2 to each epoch as a callback.
        This also changes self.verbose to 2 to prevent log print mess up.
    average_type: str (default='mean')
        Determines what to fit as the average: 'mean', or 'median'. The average is the last
        node in the output layer and does not reflect a quantile, but rather estimates the central
        tendency of the data. Setting to 'mean' results in fitting that node with MSE, and
        setting this to 'median' results in fitting that node with MAE (equal to 0.5 quantile).
    kwargs: dict, optional
        Not used. Just for compatibility with other SAM models.

    Attributes
    ----------
    feature_engineer_: Sklearn transformer
        The transformer used on the raw data before prediction
    n_inputs_: integer
        The number of inputs used for the underlying neural network
    n_outputs_: integer
        The number of outputs (columns) from the model
    prediction_cols_: array of strings
        The names of the output columns from the model.
    model_: Keras model
        The underlying keras model

    Examples
    --------
    >>> import pandas as pd
    >>> from sam.models import MLPTimeseriesRegressor
    >>> from sam.feature_engineering import SimpleFeatureEngineer
    >>> from sam.datasets import load_rainbow_beach
    ...
    >>> data = load_rainbow_beach()
    >>> X, y = data, data["water_temperature"]

    >>> simple_features = SimpleFeatureEngineer(
    ...     rolling_features=[
    ...         ("wave_height", "mean", 24),
    ...         ("wave_height", "mean", 12),
    ...     ],
    ...     time_features=[
    ...         ("hour_of_day", "cyclical"),
    ...     ],
    ...     keep_original=False,
    ... )
    >>> model = MLPTimeseriesRegressor(
    ...     predict_ahead=(0,),
    ...     feature_engineer=simple_features,
    ...     verbose=0,
    ... )
    >>> model.fit(X, y)  # doctest: +ELLIPSIS
    <keras.callbacks.History ...
    """

    def __init__(
        self,
        predict_ahead: Sequence[int] = (0,),
        quantiles: Sequence[float] = (),
        use_diff_of_y: bool = False,
        timecol: str = None,
        y_scaler: TransformerMixin = None,
        feature_engineer: BaseFeatureEngineer = None,
        n_neurons: int = 200,
        n_layers: int = 2,
        batch_size: int = 16,
        epochs: int = 20,
        lr: float = 0.001,
        dropout: float = None,
        momentum: float = None,
        verbose: int = 1,
        r2_callback_report: bool = False,
        average_type: str = "mean",
        **kwargs,
    ) -> None:
        super().__init__(
            predict_ahead=predict_ahead,
            quantiles=quantiles,
            use_diff_of_y=use_diff_of_y,
            timecol=timecol,
            y_scaler=y_scaler,
            feature_engineer=feature_engineer,
            **kwargs,
        )
        self.n_neurons = n_neurons
        self.n_layers = n_layers
        self.batch_size = batch_size
        self.epochs = epochs
        self.lr = lr
        self.momentum = momentum
        self.dropout = dropout
        self.verbose = verbose
        self.r2_callback_report = r2_callback_report
        self.average_type = average_type

        if self.average_type == "median" and 0.5 in self.quantiles:
            raise ValueError(
                "average_type is mean, but 0.5 is also in quantiles (duplicate). "
                "Either set average_type to mean or remove 0.5 from quantiles"
            )

    def get_untrained_model(self) -> Callable:
        """
        Returns a simple 2d keras model.
        This is just a wrapper for sam.models.create_keras_quantile_mlp

        Overwrites the abstract method from BaseTimeseriesRegressor
        """

        return create_keras_quantile_mlp(
            n_input=self.n_inputs_,
            n_neurons=self.n_neurons,
            n_layers=self.n_layers,
            n_target=len(self.predict_ahead),
            quantiles=self.quantiles,
            lr=self.lr,
            momentum=self.momentum,
            dropout=self.dropout,
            average_type=self.average_type,
        )

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        validation_data: Tuple[pd.DataFrame, pd.Series] = None,
        **fit_kwargs,
    ) -> Callable:
        """
        This function does the following:
        - Validate that the input is monospaced and has enough rows
        - Perform differencing on the target
        - Create feature engineer by calling `self.get_feature_engineer()`
        - Fitting/applying the feature engineer
        - Bookkeeping to create the output columns
        - Remove rows with nan that can't be used for fitting
        - Get untrained model with `self.get_untrained_model()`
        - Fit the untrained model and return the history object
        - Optionally, preprocess validation data to give to the fit
        - Pass through any other fit_kwargs to the fit function

        Overwrites the abstract method from BaseTimeseriesRegressor

        Parameters
        ----------
        X: pd.DataFrame
            The independent variables used to 'train' the model
        y: pd.Series
            Target data (dependent variable) used to 'train' the model.
        validation_data: tuple(pd.DataFrame, pd.Series) (X_val, y_val respectively)
            Data used for validation step

        Returns
        -------
        tf.keras.callbacks.History:
            The history object after fitting the keras model
        """
        (
            X_transformed,
            y_transformed,
            X_val_transformed,
            y_val_transformed,
        ) = self.preprocess_fit(X, y, validation_data)

        self.model_ = self.get_untrained_model()

        if validation_data is not None:
            validation_data = (X_val_transformed, y_val_transformed)

        if self.r2_callback_report:

            all_data = {"X_train": X_transformed, "y_train": y_transformed}

            if validation_data is not None:
                all_data["X_val"] = X_val_transformed
                all_data["y_val"] = y_val_transformed

            # append to the callbacks argument
            if "callbacks" in fit_kwargs.keys():
                # early stopping should be last callback to work properly
                fit_kwargs["callbacks"] = [
                    R2Evaluation(all_data, self.prediction_cols_, self.predict_ahead)
                ] + fit_kwargs["callbacks"]
            else:
                fit_kwargs["callbacks"] = [
                    R2Evaluation(all_data, self.prediction_cols_, self.predict_ahead)
                ]

            # Keras verbosity can be in [0, 1, 2].
            # If verbose is 1, this means that the user wants to display messages.
            # However, keras printing messes up with custom callbacks and verbose == 1,
            # see: https://github.com/keras-team/keras/issues/2354
            # The only solution for now is to set verbose to 2 when using custom callbacks.
            if self.verbose == 1:
                self.verbose = 2

        # Fit model
        history = self.model_.fit(
            X_transformed.values,
            y_transformed,
            batch_size=self.batch_size,
            epochs=self.epochs,
            verbose=self.verbose,
            validation_data=validation_data,
            **fit_kwargs,
        )
        return history

    def predict(
        self,
        X: pd.DataFrame,
        y: pd.Series = None,
        return_data: bool = False,
        force_monotonic_quantiles: bool = False,
    ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]:
        """
        Make a prediction, and undo differencing in the case it was used

        Important! This is different from sklearn/tensorflow API...
        We need y during prediction for two reasons:
        1) a lagged version is used for feature engineering
        2) The underlying model can predict a differenced number, and then we want to output the
           'real' prediction, so we need y to undo the differencing
        Keep in mind that prediction will work if you are predicting the future. e.g. you have
        data from 00:00-12:00, and are predicting 4 hours into the future, it will predict what
        the value will be at 4:00-16:00

        Overwrites the abstract method from BaseTimeseriesRegressor

        Parameters
        ----------
        X: pd.DataFrame
            The independent variables used to predict.
        y: pd.Series
            The target values
        return_data: bool, optional (default=False)
            whether to return only the prediction, or to return both the prediction and the
            transformed input (X) dataframe.
        force_monotonic_quantiles: bool, optional (default=False)
            whether to force quantiles to not overlap. When fitting multiple quantile regressions
            it is possible that individual quantile regression lines over-lap, or in other words,
            a quantile regression line fitted to a lower quantile predicts higher that a line
            fitted to a higher quantile. If this occurs for a certain prediction, the output
            distribution is invalid. We can force monotonicity by making the outer quantiles
            at least as high as the inner quantiles.

        Returns
        -------
        prediction: pd.DataFrame
            The predictions coming from the model
        X_transformed: pd.DataFrame, optional
            The transformed input data, when return_data is True, otherwise None
        """
        self.validate_data(X)

        if y is None and self.use_diff_of_y:
            raise ValueError("You must provide y when using use_diff_of_y=True")

        X_transformed = self.preprocess_predict(X, y)
        prediction = self.model_.predict(X_transformed)

        prediction = self.postprocess_predict(
            prediction, X, y, force_monotonic_quantiles=force_monotonic_quantiles
        )

        if return_data:
            return prediction, X_transformed
        else:
            return prediction

    def dump(self, foldername: Union[str, Path], prefix: str = "model") -> None:
        """
        Writes the following files:
        * prefix.pkl
        * prefix.h5

        to the folder given by foldername. prefix is configurable, and is
        'model' by default

        Overwrites the abstract method from BaseTimeseriesRegressor

        Parameters
        ----------
        foldername: str
            The name of the folder to save the model
        prefix: str, optional (Default='model')
            The name of the model
        """
        # This function only works if the estimator is fitted
        check_is_fitted(self, "model_")

        import cloudpickle

        foldername = Path(foldername)

        # TEMPORARY
        self.model_.save(foldername / (prefix + ".h5"))

        # Set the models to None temporarily, because they can't be pickled
        backup, self.model_ = self.model_, None

        with open(foldername / (prefix + ".pkl"), "wb") as f:
            cloudpickle.dump(self, f)

        # Set it back
        self.model_ = backup

    @classmethod
    def load(cls, foldername: Union[str, Path], prefix="model"):
        """
        Reads the following files:
        * prefix.pkl
        * prefix.h5

        from the folder given by foldername. prefix is configurable, and is
        'model' by default
        Output is an entire instance of the fitted model that was saved

        Overwrites the abstract method from BaseTimeseriesRegressor

        Returns
        -------
        Keras model
        """
        import cloudpickle
        from tensorflow import keras

        foldername = Path(foldername)
        with open(foldername / (prefix + ".pkl"), "rb") as f:
            obj = cloudpickle.load(f)

        loss = obj._get_loss()
        obj.model_ = keras.models.load_model(
            foldername / (prefix + ".h5"), custom_objects={"mse_tilted": loss}
        )
        return obj

    def _get_loss(self) -> Union[str, Callable]:
        """
        Convenience function, mirrors create_keras_quantile_mlp
        Only needed for loading, since it is a custom object, it is not
        saved in the .h5 file by default
        """
        if len(self.quantiles) == 0:
            return "mse"
        else:

            def mse_tilted(y, f):
                loss = keras_joint_mse_tilted_loss(y, f, self.quantiles, len(self.predict_ahead))
                return loss

        return mse_tilted

    def summary(self, print_fn: Callable = print) -> None:
        """
        Combines several methods to create a 'wrapper' summary method.

        Parameters
        ----------
        print_fn: Callable (default=print)
            A function for writting down results
        """
        # This function only works if the estimator is fitted
        check_is_fitted(self, "model_")
        print_fn(str(self))
        print_fn(self.get_feature_names_out())
        self.model_.summary(print_fn=print_fn)

    def quantile_feature_importances(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        score: Union[str, Callable] = None,
        n_iter: int = 5,
        sum_time_components: bool = False,
        random_state: int = None,
    ) -> pd.DataFrame:
        """
        Computes feature importances based on the loss function used to estimate the average.
        This function uses ELI5's `get_score_importances:
        <https://eli5.readthedocs.io/en/latest/autodocs/permutation_importance.html>`
        to compute feature importances. It is a method that measures how the score decreases when
        a feature is not available.
        This is essentially a model-agnostic type of feature importance that works with
        every model, including keras MLP models.

        Note that we compute feature importance over the average (the central trace, and the last
        output node, either median or mean depending on self.average_type), and do not include
        the quantiles in the loss calculation. Initially, the quantiles were included, but
        experimentation showed that importances behaved very badly when including the
        quantiles in the loss: importances were sometimes consistently negative (i.e. in all
        random iterations), while these features should have been important according to theory,
        and excluding them indeed lead to much worse model performance. This behavior goes away
        when only using the mean trace to estimate feature importance.

        Parameters
        ----------
        X: pd.DataFrame
            dataframe with test or train features
        y: pd.Series
            dataframe with test or train target
        score: str or function, optional (default=None)
            Either a function with signature score(X, y, model)
            that returns a scalar. Will be used to measure score decreases for ELI5.
            If None, defaults to MSE or MAE depending on self.average_type.
            Note that if score computes a loss (i.e. higher is worse), negative values indicate
            positive contribution to model performance (i.e. negative score decrease means that
            removing this feature will increase the metric, which is a bad thing with MAE/MSE).
        n_iter: int, optional (default=5)
            Number of iterations to use for ELI5. Since ELI5 results can vary wildly, increasing
            this parameter may provide more stability at the cost of a longer runtime
        sum_time_components: bool, optional (default=False)
            if set to true, sums feature importances of the different subfeatures of each time
            component (i.e. weekday_1, weekday_2 etc. in one 'weekday' importance)
        random_state: int, optional (default=None)
            Used for shuffling columns of matrix columns.

        Returns
        -------
        score_decreases: Pandas dataframe,  shape (n_iter x n_features)
            The score decreases when leaving out each feature per iteration. The larger the
            magnitude, the more important each feature is considered by the model.

        Examples
        --------
        >>> # Example with a fictional dataset with only 2 features
        >>> import pandas as pd
        >>> import seaborn
        >>> from sam.models import MLPTimeseriesRegressor
        >>> from sam.feature_engineering import SimpleFeatureEngineer
        >>> from sam.datasets import load_rainbow_beach
        ...
        >>> data = load_rainbow_beach()
        >>> X, y = data, data["water_temperature"]
        >>> test_size = int(X.shape[0] * 0.33)
        >>> train_size = X.shape[0] - test_size
        >>> X_train, y_train = X.iloc[:train_size, :], y[:train_size]
        >>> X_test, y_test = X.iloc[train_size:, :], y[train_size:]
        ...
        >>> simple_features = SimpleFeatureEngineer(
        ...     rolling_features=[
        ...         ("wave_height", "mean", 24),
        ...         ("wave_height", "mean", 12),
        ...     ],
        ...     time_features=[
        ...         ("hour_of_day", "cyclical"),
        ...     ],
        ...     keep_original=False,
        ... )
        ...
        >>> model = MLPTimeseriesRegressor(
        ...     predict_ahead=(0,),
        ...     feature_engineer=simple_features,
        ...     verbose=0,
        ... )
        ...
        >>> model.fit(X_train, y_train)  # doctest: +ELLIPSIS
        <keras.callbacks.History ...
        >>> score_decreases = model.quantile_feature_importances(
        ...     X_test[:100], y_test[:100], n_iter=3, random_state=42)
        >>> # The score decreases of each feature in each iteration
        >>> feature_importances = score_decreases.mean()
        >>> # This will show a barplot of all the score importances, with error bars
        >>> seaborn.barplot(data=score_decreases)  # doctest: +SKIP
        """
        # Model must be fitted for this method
        check_is_fitted(self, "model_")

        if len(self.predict_ahead) > 1:
            raise NotImplementedError(
                "This method is currently not implemented " "for multiple targets"
            )

        from eli5.permutation_importance import get_score_importances

        if score is None:

            def score(X, y, model=self.model_):
                if self.average_type == "median":
                    return np.mean(np.abs(y - model.predict(X)[:, -1]))
                elif self.average_type == "mean":
                    return np.mean((y - model.predict(X)[:, -1]) ** 2)

        X_transformed = self.preprocess_predict(X, y)

        if self.predict_ahead != [0]:
            y_target = make_shifted_target(y, self.use_diff_of_y, self.predict_ahead).iloc[:, 0]
        else:
            y_target = y.copy().astype(float)

        # Remove rows with missings in either of the two arrays
        missings = np.isnan(y_target) | np.isnan(X_transformed).any(axis=1)
        X_transformed = X_transformed.loc[~missings, :]
        y_target = y_target[~missings]

        # use eli5 to compute feature importances:
        base_scores, score_decreases = get_score_importances(
            score,
            X_transformed.to_numpy(),
            y_target.to_numpy(),
            n_iter=n_iter,
            random_state=random_state,
        )

        decreases_df = pd.DataFrame(score_decreases, columns=self.get_feature_names_out())

        if sum_time_components:
            for component in self.time_components:
                these_cols = [
                    c
                    for c in decreases_df.columns
                    if c.startswith("%s_%s_" % (self.timecol, component))
                ]
                decreases_df[component] = decreases_df[these_cols].sum(axis=1)
                decreases_df = decreases_df.drop(these_cols, axis=1)

        return decreases_df

    def get_explainer(
        self, X: pd.DataFrame, y: pd.Series = None, sample_n: int = None
    ) -> SamShapExplainer:
        """
        Obtain a shap explainer-like object. This object can be used to
        create shap values and explain predictions.

        Keep in mind that this will explain
        the created features from `self.get_feature_names_out()`, not the input features.
        To help with this, the explainer comes with a `test_values()` attribute
        that calculates the test values corresponding to the shap values

        Parameters
        ----------
        X: pd.DataFrame
            The dataframe used to 'train' the explainer
        y: pd.Series, optional (default=None)
            Target data used to 'train' the explainer. Only required when self.predict_ahead > 0.
        sample_n: integer, optional (default=None)
            The number of samples to give to the explainer. It is recommended that
            if your background set is greater than 5000, to sample for performance reasons.

        Returns
        -------
        SamShapExplainer:
            Custom Sam object that inherits from shap.DeepExplainer

        Examples
        --------
        >>> import pandas as pd
        >>> import shap
        >>> from sam.models import MLPTimeseriesRegressor
        >>> from sam.feature_engineering import SimpleFeatureEngineer
        >>> from sam.datasets import load_rainbow_beach
        ...
        >>> data = load_rainbow_beach()
        >>> X, y = data, data["water_temperature"]
        >>> test_size = int(X.shape[0] * 0.33)
        >>> train_size = X.shape[0] - test_size
        >>> X_train, y_train = X.iloc[:train_size, :], y[:train_size]
        >>> X_test, y_test = X.iloc[train_size:, :], y[train_size:]
        ...
        >>> simple_features = SimpleFeatureEngineer(
        ...     rolling_features=[
        ...         ("wave_height", "mean", 24),
        ...         ("wave_height", "mean", 12),
        ...     ],
        ...     time_features=[
        ...         ("hour_of_day", "cyclical"),
        ...     ],
        ...     keep_original=False,
        ... )
        ...
        >>> model = MLPTimeseriesRegressor(
        ...     predict_ahead=(0,),
        ...     feature_engineer=simple_features,
        ...     verbose=0,
        ... )
        ...
        >>> model.fit(X_train, y_train)  # doctest: +ELLIPSIS
        <keras.callbacks.History ...
        >>> explainer = model.get_explainer(X_test, y_test, sample_n=10)
        >>> shap_values = explainer.shap_values(X_test[0:10], y_test[0:10])
        >>> test_values = explainer.test_values(X_test[0:10], y_test[0:10])
        >>> shap.force_plot(explainer.expected_value[0], shap_values[0][-1,:],
        ...                 test_values.iloc[-1,:], matplotlib=True)
        """
        import shap

        X_transformed = self.preprocess_predict(X, y, dropna=True)
        if sample_n is not None:
            # Sample some rows to increase performance later
            sampled = np.random.choice(X_transformed.shape[0], sample_n, replace=False)
            X_transformed = X_transformed.iloc[sampled, :]
        explainer = shap.KernelExplainer(self.model_.predict, X_transformed.to_numpy())
        return SamShapExplainer(explainer, self, preprocess_predict=self.preprocess_predict)



from pathlib import Path
from typing import Any, Callable, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from sam.models.base_model import BaseTimeseriesRegressor
from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y


class ConstantTemplate(BaseEstimator, RegressorMixin):
    """Constant regression template for usage in BaseQuantileRegressor

    This template class follows scikit-learn estimator principles and
    always predicts quantiles and median values.

    During predict it uses the number of rows in the input to determine the
    shape of the prediction. Apart from that, the model only uses the target y
    for all calculations.

    Parameters
    ----------
    predict_ahead: Sequence of int or int, optional (default=0)
        How many timepoints we want to predict ahead. Is only used to determine the
        number of rows in the output.
    quantiles: Sequence of float, optional (default=())
        The quantiles that need to be present in the output.
    average_type: str, optional (default="median")
        The type of average that is used to calculate the median and quantiles.

    """

    def __init__(
        self,
        predict_ahead: Sequence[int] = (0,),
        quantiles: Sequence[float] = (),
        average_type: str = "median",
    ) -> None:
        self.predict_ahead = predict_ahead
        self.quantiles = quantiles
        self.average_type = average_type

    def fit(self, X: Any, y: Any, **kwargs: dict):
        """Fit the model

        The X parameter is only used so it is compatible with sklearn.
        Fits the quantiles and median values using the y data and saves
        to self.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input data, not used in this function. This parameter is only
            used for compatibility.
        y : array-like of shape (n_samples,)
            The target data that is used for determining the median and quantile
            values

        Returns
        -------
        The fitted model
        """
        X, y = check_X_y(X, y, y_numeric=True)
        self.n_features_in_ = X.shape[1]

        if self.average_type == "median":
            self.model_average_ = np.median(y)
        else:
            self.model_average_ = np.mean(y)
        self.model_quantiles_ = np.quantile(y[~np.isnan(y)], q=np.sort(self.quantiles))
        return self

    def predict(self, X: Any):
        """Predict using the model

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            input data, only used to determine the amount of rows that need
            to be present in the output.

        Returns
        -------
        array-like of shape (n_samples * predict_ahead, len(quantiles) + 1)
            output of the prediction, containing the median and and quantiles
        """
        check_is_fitted(self)
        X = check_array(X)

        prediction = np.append(self.model_quantiles_, self.model_average_)
        prediction = np.repeat(prediction, len(self.predict_ahead))
        if len(self.quantiles) > 1:
            return np.tile(prediction, (len(X), 1))
        else:
            return np.tile(prediction, len(X))

    def _more_tags(self):
        """helper function to make sure this class
        passes the check_estimator check
        """
        return {"poor_score": True}


class ConstantTimeseriesRegressor(BaseTimeseriesRegressor):
    """Constant Regression model

    Baseline model that always predict the median and quantiles.
    This model can be used as a benchmark or fall-back method, since the
    predicted median and quantiles can still be used to trigger alarms.
    Also see https://en.wikipedia.org/wiki/Statistical_process_control

    This model uses the same init parameters as the other SAM models for
    compatibility, but ignores all of the feature engineering parameters

    Note: using use_diff_of_y changes how this model works; instead of predicting
    static bounds, it will fit the median and quantiles on the differenced target
    then it will undo the differencing by adding those values to the last timestep,
    resulting in a model that predicts the last timestep + the median difference.
    This approach works especially when trying to predict a signal that has a continuous
    trend.

    Parameters
    ----------
    predict_ahead: tuple of integers, optional (default=(0,))
        how many steps to predict ahead. For example, if (1, 2), the model will predict both 1 and
        2 timesteps into the future. If (0,), predict the present. If not equal to (0,),
        predict the future. Combine with `use_diff_of_y` to get a persistence benchmark forecasting
        model.
    quantiles: tuple of floats, optional (default=())
        The quantiles to predict. Values between 0 and 1. Keep in mind that the mean will be
        predicted regardless of this parameter
    use_diff_of_y: bool, optional (default=True)
        If True differencing is used (the difference between y now and shifted y),
        else differencing is not used (shifted y is used).
    timecol: string, optional (default=None)
        If not None, the column to use for constructing time features. For now,
        creating features from a DateTimeIndex is not supported yet.
    y_scaler: object, optional (default=None)
        Should be an sklearn-type transformer that has a transform and inverse_transform method.
        E.g.: StandardScaler() or PowerTransformer().
    average_type: str = "median",
        The type of average that is used to calculate the median and quantiles.
        FOR NOW ONLY "median" IS SUPPORTED.
    kwargs: dict, optional
        Not used. Just for compatibility with other SAM models.

    Examples
    --------
    >>> from sam.models import ConstantTimeseriesRegressor
    >>> from sam.data_sources import read_knmi
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.metrics import mean_squared_error
    ...
    >>> # Prepare data
    >>> data = read_knmi('2018-02-01', '2019-10-01', latitude=52.11, longitude=5.18, freq='hourly',
    ...                  variables=['FH', 'FF', 'FX', 'T', 'TD', 'SQ', 'Q', 'DR', 'RH', 'P',
    ...                             'VV', 'N', 'U', 'IX', 'M', 'R', 'S', 'O', 'Y'])
    >>> y = data['T']
    >>> X = data.drop('T', axis=1)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y, train_size=0.8, shuffle=False)
    ...
    >>> model = ConstantTimeseriesRegressor(timecol='TIME', quantiles=[0.25, 0.75])
    ...
    >>> model.fit(X_train, y_train)
    >>> pred = model.predict(X_test, y_test)
    >>> pred.head()
           predict_lead_0_q_0.25  predict_lead_0_q_0.75  predict_lead_0_mean
    11655                   56.0                  158.0                101.0
    11656                   56.0                  158.0                101.0
    11657                   56.0                  158.0                101.0
    11658                   56.0                  158.0                101.0
    11659                   56.0                  158.0                101.0
    """

    def __init__(
        self,
        predict_ahead: Sequence[int] = (0,),
        quantiles: Sequence[float] = (),
        use_diff_of_y: bool = False,
        timecol: str = None,
        y_scaler: TransformerMixin = None,
        average_type: str = "median",
        **kwargs,
    ) -> None:
        super().__init__(
            predict_ahead=predict_ahead,
            quantiles=quantiles,
            use_diff_of_y=use_diff_of_y,
            timecol=timecol,
            y_scaler=y_scaler,
            average_type=average_type,
            **kwargs,
        )

    def get_untrained_model(self) -> Callable:
        """Returns an underlying model that can be trained

        Creates an instance of the ConstantTemplate class

        Returns
        -------
        A trainable model class
        """
        return ConstantTemplate(predict_ahead=self.predict_ahead, quantiles=self.quantiles)

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        validation_data: Tuple[pd.DataFrame, pd.Series] = None,
        **fit_kwargs,
    ) -> Callable:
        """Fit the ConstantTimeseriesRegressor model

        This function will preprocess the input data, get the untrained underlying model
        and fits the model.

        For compatibility reasons the method acceps fit_kwargs, that are not used.

        Parameters
        ----------
        X: pd.DataFrame
            The independent variables used to 'train' the model
        y: pd.Series
            Target data (dependent variable) used to 'train' the model.
        validation_data: tuple(pd.DataFrame, pd.Series) (X_val, y_val respectively)
            Data used for validation step

        Returns
        -------
        Always returns None, since there is no history object of the fit procedure
        """
        X_transformed, y_transformed, _, _ = self.preprocess_fit(X, y, validation_data)
        self.model_ = self.get_untrained_model()
        self.model_.fit(X_transformed, y_transformed)
        return None

    def predict(
        self, X: pd.DataFrame, y: pd.Series = None, return_data: bool = False, **predict_kwargs
    ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]:
        """
        Predict using the ConstantTimeseriesRegressor

        This will either predict the static bounds that were fitted during
        fit() or when using `use_diff_of_y` it will predict the last timestep plus
        the median/quantile difference.

        In the first situation X is only used to determine how many datapoints
        need to be predicted. In the latter case it will use X to undo the differencing.

        For compatibility reasons the method accepts predict_kwargs, that are not used.

        Parameters
        ----------
        X: pd.DataFrame
            The independent variables used to predict.
        y: pd.Series
            The target values
        return_data: bool, optional (default=False)
            whether to return only the prediction, or to return both the prediction and the
            transformed input (X) dataframe.

        Returns
        -------
        prediction: pd.DataFrame
            The predictions coming from the model
        X_transformed: pd.DataFrame, optional
            The transformed input data, when return_data is True, otherwise None
        """
        self.validate_data(X)

        X_transformed = self.preprocess_predict(X, y)
        prediction = self.model_.predict(X_transformed)

        prediction = self.postprocess_predict(prediction, X, y)

        if return_data:
            return prediction, X_transformed
        else:
            return prediction

    def dump(self, foldername: str, prefix: str = "model") -> None:
        """
        Writes the instanced model to foldername/prefix.pkl

        prefix is configurable, and is 'model' by default

        Overwrites the abstract method from SamQuantileRegressor

        Parameters
        ----------
        foldername: str
            The name of the folder to save the model
        prefix: str, optional (Default='model')
            The name of the model
        """
        # This function only works if the estimator is fitted
        check_is_fitted(self, "model_")

        import cloudpickle

        foldername = Path(foldername)

        with open(foldername / (prefix + ".pkl"), "wb") as f:
            cloudpickle.dump(self, f)

    @classmethod
    def load(cls, foldername, prefix="model") -> Callable:
        """
        Reads and loads the model located at foldername/prefix.pkl

        prefix is configurable, and is 'model' by default
        Output is an entire instance of the fitted model that was saved

        Overwrites the abstract method from SamQuantileRegressor

        Returns
        -------
        A fitted ConstantTimeseriesRegressor object
        """
        import cloudpickle

        foldername = Path(foldername)
        with open(foldername / (prefix + ".pkl"), "rb") as f:
            obj = cloudpickle.load(f)

        return obj



import sys
import warnings
from typing import Union

import numpy as np
import pandas as pd


def train_r2(
    true: Union[pd.Series, np.array],
    predicted: Union[pd.Series, np.array],
    benchmark: Union[float, np.array],
):
    """
    Calculation of r2 using a benchmark (often mean/median/mode of the train set). The idea is
    that calculating r2 over a an average of the test set is 'leaking' information. This is
    especially problematic if the test set is small compared to the train set. It is therefore
    more suitable to compare the model predictions to the average estimated in the train set.
    It is also possible to define an array benchmark, for example a prediction using your model
    with input a certain feature nulled (such as rainfall prediction). Then, by comparing this
    benchmark with the model prediction with a non-null rainfall prediction it's possible to
    determine that certain (for example rain) feature is important in the model and thus
    indeed improves predictions.

    Parameters
    ----------
    true: np.array or pd.Series
        Actual timeseries
    predicted: np.array or pd.Series
        Predicted timeseries
    benchmark: float or pd.Series
        Average (mean/median/mode) estimated from the train set or a timeseries to compare to

    Returns:
    -------
    r2: float
        Custom R2 with training mean/median
    """

    if len(true.shape) > 1:
        true_ravel = np.ravel(true)
        if true_ravel.size > true.shape[0]:
            raise ValueError("true argument must be 1 dimensional")

        true = true_ravel
    if isinstance(benchmark, (pd.Series, np.ndarray)) and len(benchmark.shape) > 1:
        benchmark_ravel = np.ravel(benchmark)
        if benchmark_ravel.size > benchmark.shape[0]:
            raise ValueError("benchmark argument must be 1 dimensional")
        if benchmark_ravel != true.size:
            raise ValueError("benchmark array must be same size as true array")
        benchmark = benchmark_ravel
    if len(predicted.shape) > 1:
        predicted_ravel = np.ravel(predicted)
        if predicted_ravel.size > predicted.shape[0]:
            raise ValueError("predicted argument must be 1 dimensional")
        predicted = predicted_ravel

    num = np.nansum((true - predicted) ** 2)
    denom = np.nansum((true - benchmark) ** 2)

    r2 = 1 - (num / (denom + sys.float_info.epsilon))

    return r2


def train_mean_r2(
    true: Union[pd.Series, np.array],
    predicted: Union[pd.Series, np.array],
    train_mean: float,
):
    """
    Calculation of r2 with custom mean, so you can pass the mean from the train set.
    The idea is that calculating r2 over a test set is 'leaking' information to the mean benchmark.
    This is especially problematic if the test set is small compared to the train set.
    It is therefore more suitable to compare the model predictions to the mean estimated in the
    train set.

    Parameters
    ----------
    true: np.array or pd.Series
        Actual timeseries
    predicted: np.array or pd.Series
        Predicted timeseries
    train_mean: float
        Mean estimated from the train set

    Returns:
    -------
    r2: float
        Custom R2 with training mean
    """
    warnings.warn(
        "DEPRECATED: USE train_r2(true, predicted, benchmark)" " WITH SAME PARAMETERS",
        DeprecationWarning,
    )

    return train_r2(true, predicted, train_mean)



from typing import Dict, Sequence

import numpy as np
import pandas as pd
from sam.metrics.r2_calculation import train_r2
from tensorflow.keras.callbacks import Callback


class R2Evaluation(Callback):
    def __init__(
        self,
        all_data: Dict[str, np.ndarray],
        prediction_cols: list,
        predict_ahead: Sequence[int],
    ):
        """
        Custom keras callback that computes r2 compared to the training mean.
        Computing R2 at every batch and then averaging biases r2 estimates.
        Implementing r2 as a metric is therefore not valid, as this is evaluated every batch.
        We therefore implemented it as a callback, which is only evaluated at the end of each
        epoch.

        NOTE: this should only be used with MLPTimeseriesRegressor models, not any custom keras
        model.
        NOTE-2: this function returns r2 with the keras_model.predict function. This means that
            if values are differences in MLPTimeseriesRegressor, it returns r2 for the differenced
            values. This can deviate from r2 computed over un-differenced values.

        Parameters
        ----------
        all_data: dict
            Dictionary that should include X_train and y_train at least. If validation set is
            present should also include X_val and y_val. The training sets individually should
            be numpy arrays, and should be the same as are input to the model for training.
        prediction_cols: list
            List of columns that accompany a model.predict call
        predict_ahead: integer
            Number of timesteps ahead

        """
        self.all_data = all_data
        self.prediction_cols = prediction_cols
        self.predict_ahead = predict_ahead

    def on_epoch_end(self, epoch, logs=None):
        """
        Computes r2 compared to the training mean for each predict_ahead, calculated the
        average and prints the result
        """
        if logs is None:
            logs = {}

        val = "X_val" in self.all_data.keys()

        y_hat_train = pd.DataFrame(
            data=self.model.predict(self.all_data["X_train"]),
            index=self.all_data["X_train"].index,
            columns=self.prediction_cols,
        )

        if val:
            y_hat_val = pd.DataFrame(
                data=self.model.predict(self.all_data["X_val"]),
                index=self.all_data["X_val"].index,
                columns=self.prediction_cols,
            )

        r2s = []
        r2s_val = []
        for p in self.predict_ahead:

            if len(self.predict_ahead) > 1:
                # only add the predict ahead if needed
                thiscol = "_".join(self.all_data["y_train"].columns[0].split("_")[:-2])
                thiscol += "_lead_%d" % p
            else:
                thiscol = self.all_data["y_train"].columns[0]

            these_y_train = self.all_data["y_train"].loc[:, thiscol].values
            these_y_hat_train = y_hat_train["predict_lead_%d_mean" % p].values
            train_mean = these_y_train.mean()

            if val:
                these_y_val = self.all_data["y_val"].loc[:, thiscol].values
                these_y_hat_val = y_hat_val["predict_lead_%d_mean" % p].values

            r2s.append(train_r2(these_y_train, these_y_hat_train, train_mean))

            if val:
                r2s_val.append(train_r2(these_y_val, these_y_hat_val, train_mean))

        r2 = np.mean(r2s)
        logs["r2"] = r2
        if val:
            r2_val = np.mean(r2s_val)
            logs["val_r2"] = r2_val

        if val:
            print("r2: {:.6f} - val_r2: {:.6f}".format(r2, r2_val))
        else:
            print("r2: {:.6f}".format(r2))



from typing import Union, Optional
import pandas as pd


def datetime_train_test_split(
    *arrays: Union[pd.DataFrame, pd.Series],
    datetime: str,
    datecol: Optional[str] = None,
):
    """
    Split the dataframe into train and test sets based on datetime index values

    Parameters
    ----------
    arrays : Union[pd.DataFrame, pd.Series]
        Allowed inputs are pandas dataframes or series (with datetime index)
    datetime : str
        Datetime to split the dataframe on
    datecol : str, optional
        Name of the column containing the datetime index. If not provided, the index
        of the dataframes are used.

    Returns
    -------
    list: list
        List containing the train and test splits of the input arrays

    """
    output = []
    for array in arrays:
        if isinstance(array, pd.DataFrame) or isinstance(array, pd.Series):
            if datecol is None:
                dates = array.index.to_series()
            else:
                dates = array[datecol]
            train_index = dates < datetime
            test_index = dates >= datetime
            output.append(array.loc[train_index])
            output.append(array.loc[test_index])
        else:
            raise TypeError("Input must be pandas dataframe or series")

    return output



from .benchmark import benchmark_wrapper
from .benchmark import (
    benchmark_model,
    plot_score_dicts,
    preprocess_data_for_benchmarking,
)
from .keras_templates import create_keras_autoencoder_rnn
from .keras_templates import (
    create_keras_autoencoder_mlp,
    create_keras_quantile_mlp,
    create_keras_quantile_rnn,
)

from .sam_shap_explainer import SamShapExplainer
from .linear_model import LinearQuantileRegression
from .base_model import BaseTimeseriesRegressor
from .constant_model import ConstantTimeseriesRegressor
from .lasso_model import LassoTimeseriesRegressor
from .mlp_model import MLPTimeseriesRegressor

__all__ = [
    "benchmark_wrapper",
    "benchmark_model",
    "plot_score_dicts",
    "preprocess_data_for_benchmarking",
    "create_keras_autoencoder_rnn",
    "create_keras_autoencoder_mlp",
    "create_keras_quantile_mlp",
    "create_keras_quantile_rnn",
    "SamShapExplainer",
    "LinearQuantileRegression",
    "BaseTimeseriesRegressor",
    "ConstantTimeseriesRegressor",
    "LassoTimeseriesRegressor",
    "MLPTimeseriesRegressor",
]



import pandas as pd


def sam_format_to_wide(data: pd.DataFrame, sep: str = "_"):
    """
    Converts a typical sam-format df '(TIME, ID, TYPE, VALUE)' to wide format.
    This is almost a wrapper around pd.pivot_table, although it does a few extra things.
    It removes the multiindex that would normally occur with ID + TYPE, by concatenating them
    with a separator in between.
    It also sorts the output by 'TIME', which is not guaranteed by pivot_table.

    Parameters
    ----------
    data: pd.DataFrame
        dataframe with TIME, ID, TYPE, VALUE columns
    sep: str, optional (default='_')
        separator that will be placed between ID and TYPE to create column names.

    Returns
    -------
    data_wide: pd.DataFrame
        the data, in wide format, with 1 column for every ID/TYPE combination, as well as a
        TIME column. For example, if ID is 'abc' and TYPEs are 'debiet' and 'stand', the
        created column names will be 'abc_debiet' and 'abc_stand', as well as TIME. The result
        will be sorted by TIME, ascending. The index will be a range from 0 to `nrows`.
    """
    data = pd.pivot_table(data, values="VALUE", index=["TIME"], columns=["ID", "TYPE"])
    try:
        data.columns = [
            str(x[0]) + sep + x[1]
            for x in list(
                zip(
                    data.columns.levels[0][data.columns.codes[0]],
                    data.columns.levels[1][data.columns.codes[1]],
                )
            )
        ]
    except AttributeError:
        # In pandas 0.24, 'labels' was replaced by 'codes'
        # Since we support pandas 0.23, we need to fallback on labels if needed
        data.columns = [
            str(x[0]) + sep + x[1]
            for x in list(
                zip(
                    data.columns.levels[0][data.columns.labels[0]],
                    data.columns.levels[1][data.columns.labels[1]],
                )
            )
        ]
    data = data.reset_index().sort_values(by="TIME", axis=0)
    return data


def wide_to_sam_format(
    data: pd.DataFrame, sep: str = "_", idvalue: str = "", timecol: str = "TIME"
):
    """
    Convert a wide format dataframe to sam format.
    This function has the requirement that the dataframe has a time column, of which the name
    is given by `timecol`. Furthermore, the TYPE/ID combinations should be present in value
    column names that look like ID(sep)TYPE.

    If `sep` is `None`, then all column names (except `timecol) are assumed to be TYPE, with no
    id, so id will always be set to `idvalue`.

    Columns that look like 'A(sep)B(sep)C' will be split as 'ID = A, TYPE = B(sep)C'
    Columns that look like 'A' (without sep) will be split as 'ID = idvalue, TYPE = A'

    Parameters
    ----------
    data: pd.DataFrame
        data in wide format, with time column and other column names like 'ID(sep)TYPE'
    sep: string, optional (default='_')
        the seperator that appears in column names between id and type.
    idvalue: string, optional (default='')
        the default id value that is used when a column name contains no id
    timecol: string, optional (default='TIME')
        the column name of the time column that must be present

    Returns
    -------
    df: pd.DataFrame
        the data in sam format, with columns TIME, ID, TYPE, VALUE.
    """
    data = (
        data.rename({timecol: "TIME"}, axis=1)
        .set_index("TIME")
        .unstack()
        .reset_index()
        .rename({"level_0": "TYPE", 0: "VALUE"}, axis=1)
    )
    if sep is None:
        # The data only has one id
        data["ID"] = idvalue
        return data[["TIME", "ID", "TYPE", "VALUE"]]
    id_type = data["TYPE"].str.split(sep, n=1, expand=True)
    data["ID"] = id_type[0].replace("", idvalue)
    data["TYPE"] = id_type[1]

    # Ensure that A (without sep) will be split as ID = `idvalue`, TYPE = A
    missingtypes = data["TYPE"].isnull()
    data.loc[missingtypes, "TYPE"] = data["ID"].loc[missingtypes]
    data.loc[missingtypes, "ID"] = idvalue
    # Reorder columns
    return data[["TIME", "ID", "TYPE", "VALUE"]]



import logging
from typing import Callable, Tuple, Union

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


def setdoc(func: Callable):

    func.__doc__ = """
    This documentation covers `correct_above_threshold`, `correct_below_threshold`
    and `correct_outside_range`. These three functions can be used to filter extreme
    values or fill them with a specified method. The function can correctly handle
    series with a `DatetimeIndex`, to interpolate correctly even in the case of
    measurements with a varying frequency.

    Note: this function does not affect nans. To filter/fill missing values, use `pd.fillna
    <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.fillna.html>`_
    instead.

    Parameters
    ----------
    series: A pandas series
         The series containing potential outliers
    threshold: number, (default = 1) or a tuple (default = (0,1))
        The exclusive threshold. A number for above or below, for
        `correct_outside_range` it should be a tuple
    method: string (default = "na")
        To what the threshold exceeding values should be corrected, options are:

        - If 'na', set values to `np.nan`
        - If 'previous', set values to previous non non-exceeding, non-na value
        - If 'average', linearly interpolate values using
            `pandas.DataFrame.interpolate
            <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.interpolate.html>`_,
            *might leak and requires an index*
        - If 'clip': set to the max threshold, lower/upper in case of range
        - If 'value', set to a specific value, specified in `value` parameter
        - If 'remove', removes complete row.
    value: (default = None)
        If `method` is 'value', set the threshold exceeding entry to this value

    Returns
    -------
    series: pandas series
        The original series with the threshold exceeding values corrected

    Examples
    --------
    >>> from sam.preprocessing import correct_below_threshold
    >>> import pandas as pd
    >>> data = pd.Series([0, -1, 2, 1], index = [2, 3, 4, 6])
    >>>
    >>> correct_below_threshold(data, method = "average", threshold=0)
    2    0.0
    3    1.0
    4    2.0
    6    1.0
    dtype: float64
    >>> from sam.preprocessing import correct_outside_range
    >>> import pandas as pd
    >>> data = pd.Series([0, -1, 2, 1])
    >>>
    >>> correct_outside_range(data, method = "na", threshold=(0,1))
    0    0.0
    1    NaN
    2    NaN
    3    1.0
    dtype: float64
    >>> from sam.preprocessing import correct_above_threshold
    >>> import pandas as pd
    >>> data = pd.Series([0, -1, 2, 1])
    >>>
    >>> correct_above_threshold(data, method = "remove", threshold = 1)
    0    0
    1   -1
    3    1
    dtype: int64
    """
    return func


def _fix_values(
    series: pd.Series,
    outliers: pd.Series,
    threshold: Union[float, Tuple[float, float]],
    method: str,
    value: float,
    outside_range: bool = False,
):
    """
    Internal function used to deal with extreme values in `correct_above_threshold`,
    `correct_below_threshold` and `correct_outside_range`. Read the documentation
    from those functions to get more insigns and information regarding the
    input parameters
    """
    methods = ["na", "clip", "previous", "average", "value", "remove"]
    if method not in methods:
        raise ValueError("Method {} not allowed, it must be in {}".format(method, methods))

    logging.debug(
        "Now correcting threshold"
        "threshold={}, method={}, value={}".format(threshold, method, value)
    )

    na_locations = series.isna()

    if method == "previous":
        series.loc[outliers] = np.nan
        series = series.fillna(method="ffill")
        series.loc[na_locations] = np.nan  # Recover original nans
    elif method == "average":
        series.loc[outliers] = np.nan
        # Values seems to work with time indexes as well
        series = series.interpolate(method="values")
        series.loc[na_locations] = np.nan  # Recover original nans
    elif method == "value":
        series.loc[outliers] = value
    elif method == "remove":
        series = series.loc[~outliers]
    elif method == "na":
        series.loc[outliers] = np.nan
    elif method == "clip":
        if outside_range:
            series = series.clip(lower=threshold[0], upper=threshold[1])
        else:
            series.loc[outliers] = threshold

    logger.info(
        "Correct_outside_range changed {} values using method {}".format(sum(outliers), method)
    )
    logger.info(
        "The series previously had {} missing values, now it has {}".format(
            na_locations.sum(), series.isna().sum()
        )
    )

    return series


@setdoc
def correct_above_threshold(series, threshold=1, method="na", value=None):
    outliers = series > threshold
    return _fix_values(series, outliers, threshold, method, value)


@setdoc
def correct_below_threshold(series, threshold=0, method="na", value=None):
    outliers = series < threshold
    return _fix_values(series, outliers, threshold, method, value)


@setdoc
def correct_outside_range(series, threshold=(0, 1), method="na", value=None):
    if not threshold[0] < threshold[1]:
        raise ValueError("Threshold must be a tuple (a, b) with a < b")
    outliers = (series < threshold[0]) | (series > threshold[1])
    return _fix_values(series, outliers, threshold, method, value, outside_range=True)



from typing import List

import numpy as np
import pandas as pd

COLORS = ["#2453bd", "#7497e3", "#c3d0eb", "#d1d8e8"]


def sam_quantile_plot(
    y_true: pd.DataFrame,
    y_hat: pd.DataFrame,
    title: str = None,
    y_title: str = "",
    data_prop: int = None,
    y_range: int = None,
    date_range: list = None,
    colors: list = COLORS,
    outlier_min_q: int = None,
    predict_ahead: int = 0,
    res: str = None,
    interactive: bool = False,
    outliers: np.ndarray = None,
    outlier_window: int = 1,
    outlier_limit: int = 1,
    ignore_value: float = None,
    benchmark: pd.DataFrame = None,
    benchmark_color: str = "purple",
):
    """
    Uses the output from MLPTimeseriesRegressor predict function to create a quantile prediction
    plot. It plots the actual data, the prediction, and the quantiles as shaded regions.
    The plot displays a single prediction per timepoint (e.g. made 10 timepoints ago).
    The plot can be made for any predict_ahead, and can be resampled to any time-resolution.
    The plot also highlights outliers as all true values that fall outside the `outlier_min_q`th
    quantile.

    Note: this function only works when an even number of quantiles was used in the fit procedure!

    Parameters
    ---------
    y_true: pd.DataFrame
        Pandas DataFrame with single row (the actual values) should have same index as y_hat.
    y_hat: pd.DataFrame
        Dataframe returned by the MLPTimeseriesRegressor .predict() function.
        Columns should contain at least `predict_lead_x_mean`, where x is predict ahead
        and for each quantile: `predict_lead_x_q_y` where x is the predict_ahead, and
        y is the quantile. So e.g.:
        `['predict_lead_0_q_0.25, predict_lead_0_q_0.75, predict_lead_mean']`
    title: string (default=None)
        Title for the plot.
    y_title: string (default='')
        Title to put along the yaxis (ylabel).
    data_prop: int (default=None)
        Proportion of data range to include outside true data maxima for plot range,
        if not set to None. This is only applied if y_range is None.
    y_range: list or None (default=None)
        The minimum and maximum for the y-axis if not set to None.
    date_range: list (default=None)
        The minimum and maximum date for the x-axis if not set to None.
        e.g.:['2019-11-13', '2019-11-26']
    colors: list of strings (default=['#2453bd', '#7497e3', '#c3d0eb'])
        should be a valid colorstring for each quantile (first in list is narrowest quantile etc.)
    outlier_min_q: int (default=None)
        Outlier number to use for determining 'invalid' samples.
        1 indicates the narrowest CI.
        If 3 quantiles are used in the fit procedure, this can be either 1, 2, or 3.
        In this situation, if outlier_min_q is set to 3, all true values that fall outside of the
        third quantile are highlighted in the plot.
        Cannot be used in conjunction with `outliers`.
    predict_ahead: int (default=0)
        Number of samples ago that prediction was made.
        Should be one that was included in the MLPTimeseriesRegressor fit procedure.
    res: string (default=None)
        Time resolution to resample data to. Should be interpretable by pandas resamples
        (e.g. '5min', '1D' etc.). For this to work, the data must have datetime indices.
    interactive: bool (default=False)
        Returns a matplotlib figure if False, otherwise returns a plotly figure.
    outliers: array-like (default=None)
        Alternatively to the outlier_min_q argument, you can pass a boolean array of outliers here.
        This allows the user to specify their own rules that determine whether a sample is an
        outlier or not. Should be same length as y_hat, and should either have same index,
        or no index at all (np array) or list.
        Cannot be used in conjunction with `outlier_min_q`.
    outlier_window: int (default=1)
        the window size in which at least `outlier_limit` should be outside of `outlier_min_q`
    outlier_limit: int (default=1)
        the minimum number of outliers within outlier_window to be outside of `outlier_min_q`
    ignore_value: float (default=None)
        value to ignore during resampling (e.g. 0 for pumps that often go off)
    benchmark: pd.DataFrame
        The benchmark used to determine R2 of y_hat, for example a dataframe returned by the
        MLPTimeseriesRegressor.predict() function. Columns should contain at least
        `predict_lead_x_mean`, where x is predict ahead and for each quantile: `predict_lead_x_q_y`
        where x is the predict_ahead, and y is the quantile. So e.g.:
        `['predict_lead_0_q_0.25, predict_lead_0_q_0.75, predict_lead_mean']`
    benchmark_color: string (default='purple')
        a valid colorstring for the benchmark line/scatter color

    Returns
    ------
    fig: matplotlib.pyplot.figure if interactive=False else go.Figure
    """

    if outlier_min_q is not None and outliers is not None:
        raise ValueError("outlier_min_q and outliers cannot be used simultaneously")

    if ignore_value is not None and res is None:
        raise ValueError(
            "ignore value should only be set when using resampling (res should not be None)"
        )

    if (y_title == "") and y_true.name:
        y_title = y_true.name

    # copy to make sure we don't modify the original
    y_true = y_true.copy()
    y_hat = y_hat.copy()

    # shift prediction back to match y_true
    y_hat = y_hat.shift(predict_ahead)

    # apply date range to data to speed up the rest
    if date_range is None:
        start, end = y_true.index.min(), y_true.index.max()
    else:
        start, end = date_range[0], date_range[1]
    y_true = y_true.loc[start:end]
    y_hat = y_hat.loc[start:end]

    # set ignore_values to nan so they are ignored in the resampling
    if ignore_value is None:
        ignore_timepoints = y_true.apply(lambda x: False)
    else:
        ignore_timepoints = y_true == ignore_value

    y_true.loc[ignore_timepoints] = np.nan
    y_hat.loc[ignore_timepoints] = np.nan

    # same pre-processing steps for the benchmark
    if benchmark is not None:
        benchmark = benchmark.copy()
        benchmark = benchmark.shift(predict_ahead)
        benchmark = benchmark.loc[start:end]
        benchmark.loc[ignore_timepoints] = np.nan

    # resample to desired resolution
    if res is not None:
        y_true = y_true.resample(res).mean()
        y_hat = y_hat.resample(res).mean()
        if benchmark is not None:
            benchmark = benchmark.resample(res).mean()

    # create figure
    if interactive:
        _plot_func = _interactive_quantile_plot
    else:
        _plot_func = _static_quantile_plot

    fig = _plot_func(
        y_true=y_true,
        y_hat=y_hat,
        title=title,
        y_title=y_title,
        data_prop=data_prop,
        y_range=y_range,
        date_range=date_range,
        colors=colors,
        outlier_min_q=outlier_min_q,
        predict_ahead=predict_ahead,
        outliers=outliers,
        outlier_window=outlier_window,
        outlier_limit=outlier_limit,
        benchmark=benchmark,
        benchmark_color=benchmark_color,
    )

    return fig


def _interactive_quantile_plot(
    y_true: pd.DataFrame,
    y_hat: pd.DataFrame,
    title: str = None,
    y_title: str = "",
    data_prop: int = None,
    y_range: int = None,
    date_range: list = None,
    colors: list = COLORS,
    outlier_min_q: int = None,
    predict_ahead: int = 0,
    outliers: np.ndarray = None,
    outlier_window: int = 1,
    outlier_limit: int = 1,
    benchmark: pd.DataFrame = None,
    benchmark_color: str = "purple",
):
    import plotly.graph_objs as go

    # some bookkeeping before we start plotting
    these_cols = [c for c in y_hat.columns if "predict_lead_%d_q_" % predict_ahead in c]

    # figure out how to sort the columns
    col_order = np.argsort([float(c.split("_")[-1]) for c in these_cols])

    # determine number of quantiles
    n_quants = int((len(these_cols)) / 2)

    # setup plotly figure
    fig = go.Figure()

    # loop over quantiles to create shaded regions
    for i in list(range(0, n_quants))[::-1]:
        highcol = these_cols[col_order[n_quants + i]]
        lowcol = these_cols[col_order[n_quants - 1 - i]]
        this_ci = float(highcol.split("_")[-1]) - float(lowcol.split("_")[-1])

        y_hat_dropna = y_hat.dropna()

        fig.add_trace(
            go.Scatter(
                x=y_hat_dropna.index,
                y=y_hat_dropna[highcol],
                fill=None,
                mode="lines",
                line={"width": 0.1},
                line_color=colors[i],
                name="%.5f CI" % this_ci,
                showlegend=False,
                legendgroup="%.5f CI" % this_ci,
            )
        )

        fig.add_trace(
            go.Scatter(
                x=y_hat_dropna.index,
                y=y_hat_dropna[lowcol],
                fill="tonexty",
                mode="lines",
                line={"width": 0.1},
                line_color=colors[i],
                name="%.5f CI" % this_ci,
                legendgroup="%.5f CI" % this_ci,
            )
        )

    # now draw the mean prediction and actuals
    fig.add_trace(
        go.Scatter(
            x=y_hat.index,
            y=y_hat["predict_lead_%d_mean" % predict_ahead],
            line_color=colors[0],
            name="predicted",
        )
    )

    if benchmark is not None:
        fig.add_trace(
            go.Scatter(
                x=benchmark.index,
                y=benchmark["predict_lead_%d_mean" % predict_ahead],
                line_color=benchmark_color,
                name="benchmark",
            )
        )

    fig.add_trace(go.Scatter(x=y_true.index, y=y_true, line_color="black", name="true"))

    # now plot outliers
    if outlier_min_q is not None or outliers is not None:
        # if outliers is not passed, determine them through use of outlier_min_q
        if outlier_min_q is not None and outliers is None:
            valid_low = y_hat[these_cols[col_order[n_quants - 1 - (outlier_min_q - 1)]]]
            valid_high = y_hat[these_cols[col_order[n_quants + (outlier_min_q - 1)]]]
            outliers = (y_true > valid_high) | (y_true < valid_low)
            outliers = outliers.astype(int)
            k = np.ones(outlier_window)
            outliers = (
                np.convolve(outliers, k, mode="full")[: len(outliers)] >= outlier_limit
            ).astype(bool)

        fig.add_trace(
            go.Scatter(
                x=y_true[outliers].index,
                y=y_true[outliers],
                mode="markers",
                marker={"color": "red"},
                name="outlier",
            )
        )

    # set some plot properties
    if data_prop is not None and y_range is None:
        y_range = _get_y_range(y_true, data_prop)

    if title is not None:
        fig.layout.update(title=title)
    if y_range is not None:
        fig.layout.update(yaxis_range=y_range)
    if date_range is not None:
        fig.layout.update(xaxis_range=date_range)

    fig.layout.update(
        yaxis_title=y_title,
        width=1000,
        height=500,
    )

    return fig


def _static_quantile_plot(
    y_true: pd.DataFrame,
    y_hat: pd.DataFrame,
    title: str = None,
    y_title: str = "",
    data_prop: int = None,
    y_range: int = None,
    date_range: list = None,
    colors: list = COLORS,
    outlier_min_q: int = None,
    predict_ahead: int = 0,
    outliers: np.ndarray = None,
    outlier_window: int = 1,
    outlier_limit: int = 1,
    benchmark: pd.DataFrame = None,
    benchmark_color: str = "purple",
):
    import matplotlib.pyplot as plt
    import seaborn as sns

    # some bookkeeping before we start plotting
    these_cols = [c for c in y_hat.columns if "predict_lead_%d_q_" % predict_ahead in c]

    # figure out how to sort the columns
    col_order = np.argsort([float(c.split("_")[-1]) for c in these_cols])

    # determine number of quantiles
    n_quants = int((len(these_cols)) / 2)

    # setup plotly figure
    fig = plt.figure(figsize=(15, 7.5))

    # loop over quantiles to create shaded regions
    for i in list(range(0, n_quants))[::-1]:
        highcol = these_cols[col_order[n_quants + i]]
        lowcol = these_cols[col_order[n_quants - 1 - i]]
        this_ci = float(highcol.split("_")[-1]) - float(lowcol.split("_")[-1])

        plt.fill_between(
            y_hat.index,
            y_hat[lowcol],
            y_hat[highcol],
            alpha=0.25,
            color=colors[i],
            label="%.5f CI" % this_ci,
        )

    # now draw the mean prediction and actuals
    plt.plot(
        y_hat.index,
        y_hat["predict_lead_%d_mean" % predict_ahead],
        color=colors[0],
        label="predicted",
    )

    if benchmark is not None:
        plt.plot(
            benchmark.index,
            benchmark["predict_lead_%d_mean" % predict_ahead],
            color=benchmark_color,
            label="benchmark",
        )

    plt.plot(y_true.index, y_true, color="black", label="true")

    # now plot outliers
    if outlier_min_q is not None or outliers is not None:
        # if outliers is not passed, determine them through use of outlier_min_q
        if outlier_min_q is not None and outliers is None:
            valid_low = y_hat[these_cols[col_order[n_quants - 1 - (outlier_min_q - 1)]]]
            valid_high = y_hat[these_cols[col_order[n_quants + (outlier_min_q - 1)]]]
            outliers = (y_true > valid_high) | (y_true < valid_low)
            outliers = outliers.astype(int)
            k = np.ones(outlier_window)
            outliers = (
                np.convolve(outliers, k, mode="full")[: len(outliers)] >= outlier_limit
            ).astype(bool)

        plt.plot(
            y_true[outliers].index,
            y_true[outliers],
            "o",
            ms=5,
            color="r",
            label="outlier",
        )

    # set some plot properties
    if data_prop is not None and y_range is None:
        y_range = _get_y_range(y_true, data_prop)

    if title is not None:
        plt.title(title)
    plt.legend(loc="best")
    plt.ylabel(y_title)
    sns.despine(offset=10)
    if y_range is not None:
        plt.ylim(y_range)
    if date_range is not None:
        plt.xlim(pd.to_datetime(date_range))

    return fig


def _get_y_range(y: pd.DataFrame, data_prop: int) -> List[float]:
    """Calculates the proportional range of `y` given `data_prop`

    Parameters
    ----------
    y : pd.DataFrame
        Pandas DataFrame with single row
    data_prop: int
        Proportion of data range to include outside data maxima for plot range

    Returns
    -------
    List[float]
        List of length 2, with first index the minimum and second index the maximum of the
        proportional range of `y`
    """
    data_range = y.max() - y.min()
    ymax = y.max() + data_range * data_prop
    ymin = y.min() - data_range * data_prop
    y_range = [ymin, ymax]

    return y_range



from typing import Any

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.exceptions import NotFittedError


class SPEITransformer(BaseEstimator, TransformerMixin):
    """Standardized Precipitation (and Evaporation) Index

    Computation of standardized metric that measures relative drought
    or precipitation shortage.

    SP(E)I is a metric computed per day. Therefore daily weather data
    is required as input. This class assumes that the data contains
    precipitation columns 'RH' and optionally evaporation column 'EV24'.
    These namings are KNMI standards.

    The method computes a rolling average over the precipitation (and
    evaporation). Based on historic data (at least 30 years) the mean
    and standard deviation of the rolling average are computed across years.
    The daily rolling average is then transformed to a Z-score, by dividing
    by the corresponding mean and standard deviation.

    Smoothing can be applied to make the model more robust, and able to
    compute the SP(E)I for leap year days. If ``smoothing=False``, the
    transform method can return NA's

    The resulting score describes how dry the weather is. A very low score
    (smaller than -2) indicates extremely dry weather. A high score (above 2)
    indicates very wet weather.

    See:
    http://www.droogtemonitor.nl/index.php/over-de-droogte-monitor/theorie

    Parameters
    ----------
    metric: {"SPI", "SPEI"}, default="SPI"
        The type of KPI to compute
        "SPI" computes the Standardized Precipitation Index
        "SPEI" computed the Standardized Precipitation Evaporation Index
    window: str or int, default='30D'
        Window size to compute the rolling precipitation or precip-evap sum
    smoothing: boolean, default=True
        Whether to use smoothing on the estimated mean and std for each day of
        the year.
        When ``smoothing=True``, a centered rolling median of five steps is
        applied to the models estimated mean and standard deviations per day.
        The model definition will therefore be more robust.
        Smoothing causes less sensitivity, especially for the std.
        Use the ``plot`` method to visualize the estimated mean and std
    min_years: int, default=30
        Minimum number of years for configuration. When setting less than 30,
        make sure that the estimated model makes sense, using the ``plot``
        method
    model_: dataframe, default=None
        Ignore this variable, this is required to keep the model configured
        when creating a new instance (common in for example cross validation)

    Examples
    ----------
    >>> from sam.data_sources import read_knmi
    >>> from sam.feature_engineering import SPEITransformer
    >>> knmi_data = read_knmi(start_date='1960-01-01', end_date='2020-01-01',
    ...     variables=['RH', 'EV24'], freq='daily').set_index('TIME').dropna()
    >>> knmi_data['RH'] = knmi_data['RH'].divide(10).clip(0)
    >>> knmi_data['EV24'] = knmi_data['EV24'].divide(10)
    >>> spi = SPEITransformer().configure(knmi_data)
    >>> spi.transform(knmi_data)  # doctest: +ELLIPSIS
                SPEI_30D
    TIME ...
    """

    def __init__(
        self,
        metric: str = "SPEI",
        window: str = "30D",
        smoothing: bool = True,
        min_years: int = 30,
        model_: pd.DataFrame = None,
    ):
        self.window = window
        self.metric = metric
        self.smoothing = smoothing
        self.min_years = min_years
        self.model_ = model_

    def _check_configured(self) -> None:
        if self.model_ is None:
            raise NotFittedError("model_ is None, call `configure` first")

    def _check_input(self, X: pd.DataFrame) -> None:
        """Check if required columns are present in dataframe `X`

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe to check the columns of
        """
        if "RH" not in X.columns:
            raise ValueError("Dataframe X should contain columns 'RH'")
        if ("EV24" not in X.columns) and (self.metric == "SPEI"):
            raise ValueError("Metric SPEI requires X to have 'EV24' column")

    def _compute_target(self, X: pd.DataFrame) -> pd.Series:
        """Compute target depending on metric type

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe to extract target from

        Returns
        -------
        pd.Series
            target series
        """
        if self.metric == "SPI":
            target = X["RH"]
        elif self.metric == "SPEI":
            target = X["RH"] - X["EV24"]
        else:
            raise ValueError("Invalid metric type, choose either 'SPI' or 'SPEI'")
        target = target.rolling(self.window).mean()

        return target

    def configure(self, X: pd.DataFrame, y: Any = None):
        """Fit normal distribution on rolling precipitation (and evaporation)
        Apply this to historic data of precipitation (at least ``min_years`` years)

        Parameters
        ----------
        X: pandas dataframe
            A data frame containing columns 'RH' (and optionally 'EV24')
            and should have a datetimeindex
        y: Any, default=None
            Not used
        """
        self._check_input(X)
        target = self._compute_target(X)

        self._metric_name = self.metric + "_" + self.window
        self._axis_name = "Precip-Evap" if self.metric == "SPEI" else "Precip"

        results = pd.DataFrame(
            {
                self._metric_name: target,
                "month": target.index.month,
                "day": target.index.day,
            },
            index=target.index,
        )

        self.model_ = (
            results.groupby(["month", "day"])[self._metric_name]
            .agg(["count", "mean", "std"])
            .reset_index()
            .sort_values(by=["month", "day"])
        )

        n_years = self.model_["count"].max()
        # Make sure that for at least one day there are ``min_years`` samples
        if n_years < self.min_years:
            raise ValueError(
                f"Provided weather data contains less than"
                f"{self.min_years} years. "
                f"Please provide more data for configuration"
            )

        # Each day should at least have data for 50%
        # of all years in the data, otherwise estimated mean and std
        # are set to nan. This removes leap year days
        self.model_.loc[(self.model_["count"] < (n_years / 2)), ["mean", "std"]] = np.nan
        if self.smoothing:
            # To remove spikes in the mean and std
            # and create a smooth model over the year
            # smoothing is applied. This approach of 5-step median
            # is just a first approach, and does the essential trick
            # Default SP(E)I from literature does not use smoothing
            self.model_["mean"] = (
                self.model_["mean"].rolling(5, center=True, min_periods=1).median(skipna=True)
            )
            self.model_["std"] = (
                self.model_["std"].rolling(5, center=True, min_periods=1).median(skipna=True)
            )

        return self

    def fit(self, X: pd.DataFrame, y: Any = None):
        """Fit function. Does nothing other than checking input, but is required for a
        transformer. This function wil not change the SP(E)I model. The SP(E)I should be configured
        with the ``configure`` method. In this way, the ``SPEITransfomer`` can be used within a
        sklearn pipeline, without requiring > 30 years of data.

        Parameters
        ----------
        X: pandas dataframe
            A data frame containing columns 'RH' (and optionally 'EV24')
            and should have a datetimeindex
        y: Any, default=None
            Not used
        """
        self._check_configured()
        self._check_input(X)
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Transforming new weather data to SP(E)I metric

        Parameters
        ----------
        X : pd.DataFrame
            New weather data to transform

        Returns
        -------
        pd.DataFrame
            Returns a dataframe with single columns
        """
        self._check_configured()
        self._check_input(X)
        target = self._compute_target(X)

        results = pd.DataFrame(
            {
                self._metric_name: target,
                "month": target.index.month,
                "day": target.index.day,
            },
            index=target.index,
        )

        results = results.merge(
            self.model_, left_on=["month", "day"], right_on=["month", "day"], how="left"
        )
        results.index = target.index
        results[self._metric_name] = (results[self._metric_name] - results["mean"]) / results[
            "std"
        ]
        return results[[self._metric_name]]

    def plot(self):
        """Plot model
        Visualization of the configured model. This function shows the
        estimated mean and standard deviation per day of the year.
        """
        self._check_configured()
        import matplotlib.pyplot as plt

        fig, axs = plt.subplots(1, 2, figsize=(20, 5))
        axs[0].plot(self.model_["mean"])
        axs[0].set_ylabel("Mean of " + self._axis_name)
        axs[0].set_xlabel("Day of the year")
        axs[1].plot(self.model_["std"])
        axs[1].set_ylabel("Standard deviation of " + self._axis_name)
        axs[1].set_xlabel("Day of the year")



import configparser
import datetime
import logging
import re
import warnings
from os.path import isdir


warnings.filterwarnings(
    "always", category=DeprecationWarning, module=r"^{0}\.".format(re.escape(__name__))
)

# Only log if the directory exists, stops errors on unit tests
if isdir("logs"):
    # We take only the message from the sam package, not matplotlib etc.
    logger = logging.getLogger("sam")
    fh = logging.FileHandler("logs/sam_" + datetime.datetime.now().strftime("%Y-%m-%d") + ".log")
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(message)s")
    fh.setFormatter(formatter)
    logger.addHandler(fh)
    logger.setLevel(logging.INFO)

config = configparser.ConfigParser()
config.read(".config")

__all__ = [
    "data_sources",
    "datasets",
    "exploration",
    "feature_engineering",
    "logging_functions",
    "metrics",
    "models",
    "preprocessing",
    "utils",
    "validation",
    "visualization",
]



from typing import Union

from sam.validation import RemoveExtremeValues, RemoveFlatlines
from sklearn import (
    __version__ as sklearn_version,
)  # For testing if IterativeImputer is available
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import Pipeline


def create_validation_pipe(
    cols: list,
    rollingwindow: Union[int, str],
    remove_extreme_values: bool = True,
    remove_flatlines: bool = True,
    impute_values: bool = True,
    madthresh: int = 15,
    flatwindow: int = 2,
    max_iter: int = 10,
    n_nearest_features: int = 10,
    impute_method: str = "iterative",
):
    """
    Sets up a pipeline to do data validation. Can incorporate:

    - remove extreme values
    - remove flatlines
    - impute values

    Parameters
    ---------
    cols: list of strings
        which columns in the dataframe to apply data validation to
    rollingwindow: int or str
        parameter used in RemoveExtremeValues, see :ref:`RemoveExtremeValues`
    remove_extreme_values: bool (default=True)
        if set to True, will find extreme values and set to nan
    remove_flatlines: bool (default=True)
        if set to True, will find flatline signals
    impute_values: bool (default=True)
        if set to True, will impute found nan signals
    madthresh: int (default=15)
        parameter used in RemoveExtremeValues, see :ref:`RemoveExtremeValues`
    flatwindow: int (default=2)
        parameter used in RemoveFlatlines, see :ref:`RemoveFlatlines`
    max_iter: int (default=10)
        how many iterations to try for iterative_imputer
        see https://scikit-learn.org/stable/modules/generated/sklearn.impute.IterativeImputer.html
    n_nearest_features: int (default=10)
        how many close features to use for iterative_imputer
        `Example <https://scikit-learn.org/stable/modules/generated/
        sklearn.impute.IterativeImputer.html>`
    impute_method: string (default='iterative')
        if set to 'iterative', will impute values using IterativeImputer. This is much slower,
        but also much more accurate. Can also be set to any of the SimpleImputer strategies:
        'mean', 'median', 'most_frequent', 'constant'

    Returns
    -------
    pipe: sklearn.pipeline.Pipeline instance
        The input data should be a wide-format dataframe, where rows are time and columns are
        features.
        The rows of the data should be linearly increasing in time, and can contain gaps in time.
        However, when a string is used to specify the rollingwindow parameter, the input data
        should have a datetime index.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from sam.validation import create_validation_pipe
    >>> from sam.visualization import diagnostic_extreme_removal, diagnostic_flatline_removal
    >>>
    >>> # create some data
    >>> np.random.seed(10)
    >>> base = np.random.randn(100)
    >>> X_train = pd.DataFrame(np.tile(base, (3, 3)).T, columns=['1', '2', '3'])
    >>> X_test = pd.DataFrame(np.tile(base, (3, 1)).T, columns=['1', '2', '3'])
    >>> y_test = pd.Series(base, name='target')
    >>> y_train = pd.Series(np.tile(base, 3).T, name='target')
    >>>
    >>> # add outliers to y_test:
    >>> y_test.iloc[[5, 10, 61]] *= 30
    >>> # add flatlines to y_train and y_test:
    >>> y_test.iloc[20:40] = 1
    >>> y_train.iloc[20:50] = 1
    >>>
    >>> # setup pipeline
    >>> pipe = create_validation_pipe(cols=list(X_train.columns) + ['target'], rollingwindow=5,
    ...                              impute_method='iterative')
    >>>
    >>> # put data together
    >>> train_data = X_train.join(y_train)
    >>> test_data = X_test.join(y_test)
    >>>
    >>> # now fit the pipeline on the train data and transform both train and test
    >>> train_data = pd.DataFrame(pipe.fit_transform(train_data), columns=train_data.columns,
    ...                           index=train_data.index)  # doctest: +ELLIPSIS
    [IterativeImputer] ...
    >>> test_data = pd.DataFrame(pipe.transform(test_data), columns=test_data.columns,
    ...                          index=test_data.index)  # doctest: +ELLIPSIS
    [IterativeImputer] ...
    >>>
    >>> # the fitted pipeline can now be passed to diagnostic plot functions:
    >>> # create validation visualizations
    >>> f_ext = diagnostic_extreme_removal(
    ...     pipe['extreme'], test_data, 'target')
    >>> f_ext = diagnostic_flatline_removal(
    ...     pipe['flat'], test_data, 'target')
    """
    methods = ["iterative", "mean", "median", "most_frequent", "constant"]
    if impute_method not in methods:
        raise ValueError(f"impute method not in {methods}")
    if impute_method == "iterative" and sklearn_version < "0.21":
        raise EnvironmentError(
            "For iterative impute method, " "sklearn version at least 0.21 is required."
        )

    estimators = []
    if remove_extreme_values:
        REV = RemoveExtremeValues(cols=cols, rollingwindow=rollingwindow, madthresh=madthresh)
        estimators.append(["extreme", REV])

    if remove_flatlines:
        RFV = RemoveFlatlines(cols=cols, window=flatwindow)
        estimators.append(["flat", RFV])

    if impute_values:
        if impute_method == "iterative":
            # This is a experimental feature in sklearn, only import when needed
            from sklearn.experimental import enable_iterative_imputer  # noqa:F401
            from sklearn.impute import IterativeImputer

            IMP = IterativeImputer(
                estimator=LinearRegression(),
                verbose=2,
                n_nearest_features=n_nearest_features,
                max_iter=max_iter,
                random_state=42,
            )
        else:
            IMP = SimpleImputer(strategy=impute_method)
        estimators.append(["impute", ColumnTransformer([("imputer", IMP, cols)])])

    pipe = Pipeline(estimators)

    return pipe



from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from sam.feature_engineering import BuildRollingFeatures, decompose_datetime
from sam.feature_engineering.base_feature_engineering import IdentityFeatureEngineer
from sam.utils import has_strictly_increasing_index
from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin
from sklearn.compose import ColumnTransformer
from sklearn.ensemble import RandomForestRegressor
from sklearn.impute import SimpleImputer
from sklearn.linear_model import BayesianRidge, LinearRegression
from sklearn.metrics import r2_score
from sklearn.model_selection import RandomizedSearchCV, TimeSeriesSplit
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted

INPUT_VALIDATION_ERR_MESSAGE = (
    "all input must be linearly increasing in time and have a datetime index"
)


class AutomaticRollingEngineering(BaseEstimator, TransformerMixin):
    """
    Steps for automatic rolling engineering:
    - setup self.n_rollings number of different rolling features (unparameterized yet) in
        sklearn ColumnTransformer pipeline
    - find the best parameters for each of the rolling features using random search
    - setup a ColumnTransformer with these best features that can be used in the transform method

    Parameters
    ----------
    window_sizes: list of lists
        each list should be integers or one of one of scipy.stats.distributions that convert to
        a window_size for BuildRollingFeatures.
        Each sublist corresponds to range tried for the n_rollings, and should be non-overlapping.
        So if you want 2 rollings to be generated per rolling_type and per feature, this could be:
        [scipy.stats.randint(1, 24), scipy.stats.randint(24, 168)].
        Note that using long lists results in overflow error, therefore randint is recommended.
    rolling_types: list of strings (default=['mean', 'lag'])
        rolling_types to try for BuildRollingFeatures.
        Note: cannot be 'ewm'.
    n_iter_per_param: int (default=25)
        number of random values to try for each parameter. The total number of iterations is
        given by n_iter_per_param * len(window_sizes) * len(rolling_types)
    cv: int (default=3)
        number of cross-validated tries to attempt for each parameter combination
    estimator_type: str (default='lin')
        type of estimator to determine rolling importance. Can be one of: ['rf', 'lin', 'bayeslin']
    passthrough: bool (default=True)
        whether to pass original features in the transform method or not
    cyclicals: list (default=None)
        A list of pandas datetime properties, such as ['minute', 'hour', 'dayofweek', 'week'],
        that will be converted to cyclicals.
        The rationale here is that if time features are not added, the rolling engineering
        will find values for instance of 1 day ago to predict well, while actually this is simply
        a recurring daily pattern that can be captured by time features.
        Note that if timefeatures are added, they are not added in the transform method. Therefore,
        you will have to add them yourself during subsequent model building stages.
    onehots: list (default=None)
        A list of pandas datetime properties, such as ['minute', 'hour', 'dayofweek', 'week'],
        that will be converted using onehot encoding.
        The rationale here is that if time features are not added, the rolling engineering
        will find values for instance of 1 day ago to predict well, while actually this is simply
        a recurring daily pattern that can be captured by time features.
        Note that if timefeatures are added, they are not added in the transform method. Therefore,
        you will have to add them yourself during subsequent model building stages.

    Attributes
    ----------
    feature_importances_: pandas dataframe
        With 'feature_name' column and
        if estimator_type is set to 'rf', a 'coefficients' column
        if estimator_type is set to 'lin', an 'importances' column
    feature_names_: list of strings
        names of all features, depends on self.passthrough
    rolling_feature_names_: list of strings
        names of rolling features that were added

    Examples
    --------
    >>> from sam.data_sources import read_knmi
    >>> import matplotlib.pyplot as plt
    >>> import seaborn as sns
    >>> import pandas as pd
    >>> from sam.feature_engineering import AutomaticRollingEngineering
    >>> from sklearn.model_selection import train_test_split
    >>> from scipy.stats import randint

    >>> # load some data
    >>> data = read_knmi(
    ...     '2018-01-01',
    ...     '2019-01-01',
    ...     variables = ['T', 'FH', 'FF', 'FX', 'SQ', 'Q', 'DR', 'RH']).set_index(['TIME'])

    >>> # let's predict temperature 12 hours into the future
    >>> target = 'T'
    >>> fut = 12
    >>> y = data[target].shift(-fut).iloc[:-fut]
    >>> X = data.iloc[:-fut]
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y, shuffle=False)

    >>> # do the feature selection, first try without adding timefeatures
    >>> ARE = AutomaticRollingEngineering(
    ...     window_sizes=[randint(1,24)], rolling_types=['mean', 'lag'])
    >>> ARE = ARE.fit(X_train, y_train)  # doctest: +ELLIPSIS
    Fitting ...

    >>> # check some diagnostics, note: results may vary as it is a random search
    >>> r2_base, r2_rollings, yhat_base, yhat_roll = ARE.compute_diagnostics(
    ...     X_train, X_test, y_train, y_test)
    >>> print(r2_base, r2_rollings)
    0.34353081601421964 0.7027068150592539

    >>> # you can also inspect feature importances:
    >>> barplot = sns.barplot(data=ARE.feature_importances_, y='feature_name', x='coefficients')

    >>> # and make plot of the timeseries:
    >>> timeseries_fig = plt.figure(figsize=(12, 6))
    >>> timeseries_fig = plt.plot(X_test.index, y_test.ravel(), 'ok', label='data')
    >>> timeseries_fig = plt.plot(
    ...     X_test.index, yhat_base, lw=3, alpha=0.75, label='yhat_base (r2: %.2f)'%r2_base
    ... )
    >>> timeseries_fig = plt.plot(
    ...     X_test.index, yhat_roll, lw=3, alpha=0.75, label='yhat_rolling (r2: %.2f)'%r2_rollings)
    >>> timeseries_fig = plt.legend(loc='best')
    """

    def __init__(
        self,
        window_sizes: List[List],
        rolling_types: List[str] = ["mean", "lag"],
        n_iter_per_param: int = 25,
        cv: int = 3,
        estimator_type: str = "lin",
        passthrough: bool = True,
        cyclicals: Optional[List[str]] = None,
        onehots: Optional[List[str]] = None,
    ):

        self.window_sizes = window_sizes
        self.rolling_types = rolling_types
        self.n_iter_per_param = n_iter_per_param
        self.cv = cv
        self.estimator_type = estimator_type
        self.passthrough = passthrough

        if cyclicals is None:
            self.cyclicals = []
        else:
            self.cyclicals = cyclicals

        if onehots is None:
            self.onehots = []
        else:
            self.onehots = onehots

    def _setup_estimator(self) -> RegressorMixin:
        if self.estimator_type == "rf":
            estimator = RandomForestRegressor(n_estimators=100, min_samples_split=5)
        elif self.estimator_type == "lin":
            estimator = LinearRegression()
        elif self.estimator_type == "bayeslin":
            estimator = BayesianRidge()

        return estimator

    def _setup_pipeline(
        self,
        original_features: List[str],
        time_features: List[str],
    ) -> Pipeline:
        """
        Create the pipeline that is later fitted. Includes:
        - BuildRollingFeatures
        - SimpleImputer
        - Model (depending on self.estimator_type)

        Parameters
        ----------
        original_features : list of strings
            list of original feature names
        time_features : list of strings
            list of time feature names

        Returns
        -------
        Pipeline
            resulting pipeline that can be fitted
        """

        rolls = []
        for feature in original_features:
            for i in range(len(self.window_sizes)):
                # only keep originals once to prevent doubling
                rolls.append(
                    (
                        "%s_%d" % (feature, i),
                        BuildRollingFeatures(keep_original=(i == 0)),
                        [feature],
                    )
                )

        for time_feature in time_features:
            rolls.append((time_feature, IdentityFeatureEngineer(), [time_feature]))

        pipeline = Pipeline(
            steps=[
                ("rollpipe", ColumnTransformer(rolls, n_jobs=-1)),
                ("imputer", SimpleImputer()),
                ("model", self._setup_estimator()),
            ]
        )

        return pipeline

    def _setup_rolling_gridsearch_params(
        self, original_features: List[str]
    ) -> Dict[str, List[Union[str, List]]]:
        """
        Create grid of rolling feature parameters

        Parameters
        ----------
        original_features : list of strings
            list of feature names

        Returns
        -------
        dict
            dictionary with entry for each window_size and rolling_type combination for each
            original feature
        """

        param_grid = {}
        for i in range(len(self.window_sizes)):
            for feature in original_features:
                param_grid["rollpipe__%s_%d__window_size" % (feature, i)] = self.window_sizes[i]
                param_grid["rollpipe__%s_%d__rolling_type" % (feature, i)] = self.rolling_types

        return param_grid

    def _validate_params(self, X_shape: tuple) -> None:
        """
        Validate the self parameters

        Parameters
        ----------
        X_shape : tuple
            tuple of array dimensions
        """

        for window_size in self.window_sizes:
            if type(window_size) is list:
                interval: list = [np.min(window_size), np.max(window_size)]
            else:
                # otherwise we assume it is a scipy.stats distribution
                interval: tuple = window_size.interval(1)

            if interval[0] < 0:
                raise ValueError("window_size must be greater than 0, but is %d" % interval[0])

            if X_shape[0] <= interval[1]:
                raise ValueError(
                    "number of samples too small for maximum window_size "
                    + "(should be smaller than X.shape[0])"
                )

        if "ewm" in self.rolling_types:
            raise ValueError("rolling_type cannot be ewm")
        if self.n_iter_per_param <= 0:
            raise ValueError("n_iter_per_param must be greater than 0")
        if self.cv <= 1:
            raise ValueError("cv must be greater than 1")

    def _add_time_features(self, X: pd.DataFrame) -> Tuple[pd.DataFrame, List[str]]:
        """
        Add the timefeatures given in 'components' to X

        Parameters
        ----------
        X: pandas dataframe
            with time index and other features as columns

        Returns
        -------
        X: pandas dataframe
            original X dataframe with the time features added
        timecols: list of strings
            list of time feature names
        """

        components: list = self.cyclicals + self.onehots
        if len(components) > 0:
            times_df = pd.DataFrame({"TIME": X.index})
            time_features: pd.DataFrame = decompose_datetime(
                times_df,
                components=components,
                onehots=self.onehots,
                cyclicals=self.cyclicals,
                keep_original=True,
            )
            time_features = time_features.set_index("TIME", drop=True)

            # some columns can be constant (i.e. higher res than data is in). However, removing
            # all constant columns will also remove one-hot-encoded features that are not
            # currently in the dataset (like week_52 if you only have partial year data).
            # So, we simply leave all constant features in there.

            X = X.join(time_features)
            timecols: list = time_features.columns
        else:
            timecols = []

        return X, timecols

    def get_feature_names_out(self, input_features=None) -> List[str]:
        check_is_fitted(self, "feature_names_")
        return self.feature_names_

    def fit(self, X: pd.DataFrame, y: pd.DataFrame):
        """
        Finds the best rolling feature parameters and sets up transformer for the transform method
        Note!: all input must be linearly increasing in time and have a datetime index.

        Parameters
        ----------
        X: pandas dataframe
            with shape [n_samples x n_features]
        y: pandas dataframe
            with shape [n_samples]
        """

        if not has_strictly_increasing_index(X):
            raise ValueError(INPUT_VALIDATION_ERR_MESSAGE)
        if not has_strictly_increasing_index(y):
            raise ValueError(INPUT_VALIDATION_ERR_MESSAGE)

        # save feature names here before adding timefeatures
        original_features: list = X.columns

        # add time time features
        X, timefeatures = self._add_time_features(X)

        self._validate_params(np.shape(X))

        # setup pipeline
        pipeline: Pipeline = self._setup_pipeline(original_features, timefeatures)
        param_grid: dict = self._setup_rolling_gridsearch_params(original_features)

        n_iter: int = self.n_iter_per_param * len(self.window_sizes) * len(self.rolling_types)

        # find best rollings using randomsearch (future improvement: use BayesSearchCV from skopt)
        search = RandomizedSearchCV(
            pipeline,
            param_distributions=param_grid,
            n_jobs=-1,
            cv=TimeSeriesSplit(self.cv),
            verbose=2,
            n_iter=n_iter,
        )
        search.fit(X, y)

        # recover feature names
        feature_names: list = search.best_estimator_["rollpipe"].get_feature_names_out()

        # fix names from e.g. 'RH_1__RH#mean_96' to 'RH#mean_96'
        feature_names: list = [f.split("__")[1] for f in feature_names]
        self.rolling_feature_names_: list = [f for f in feature_names if "#" in f]

        # convert these names to a set of BuildRollingFeatures that creates these features
        transformers = []
        for name in self.rolling_feature_names_:
            feature_name: str = name.split("#")[0]
            rolling_type: str = name.split("#")[1].split("_")[0]
            window_size = int(name.split("_")[-1])

            transformers.append(
                [
                    "%s_%s_%d" % (feature_name, rolling_type, window_size),
                    BuildRollingFeatures(
                        rolling_type=rolling_type,
                        window_size=window_size,
                        keep_original=False,
                    ),
                    [feature_name],
                ]
            )

        # now save the columntransformer to the self object to be used in the transform method
        self._transformer = ColumnTransformer(transformers, n_jobs=-1)
        self._transformer = self._transformer.fit(X)

        _ = search.best_estimator_[:-1].transform(X)
        if self.estimator_type == "rf":
            imp_name = "importances"
            importances: np.ndarray = search.best_estimator_["model"].feature_importances_
        elif self.estimator_type == "lin":
            imp_name = "coefficients"
            importances: np.ndarray = search.best_estimator_["model"].coef_

        # also save the feature importances under convenient name:
        self.feature_importances_ = pd.DataFrame(
            {imp_name: importances, "feature_name": feature_names}
        ).sort_values(by=imp_name, ascending=False)

        # this is required for setup of df including timefeatures in transform
        if self.passthrough:
            self.feature_names_full_: list = feature_names
        else:
            self.feature_names_full_: list = self.rolling_feature_names_

        # these are the final feature names
        self.timecols: list = [f for f in self.feature_names_full_ if "TIME" in f]
        self.feature_names_: list = [f for f in self.feature_names_full_ if "TIME" not in f]

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Applies the BuildRollingFeature transformations found to work best in the fit method
        Note!: all input must be linearly increasing in time and have a datetime index.

        Parameters
        ----------
        X: pandas dataframe
            with shape [n_samples x n_features]

        Returns
        -------
        X_transformed: pandas DataFrame
            with shape [n_samples x n_features]
        """

        if not has_strictly_increasing_index(X):
            raise ValueError(INPUT_VALIDATION_ERR_MESSAGE)

        check_is_fitted(self, "feature_names_")

        # we need to add time time features for the transformer to work
        X, _ = self._add_time_features(X)

        # create the rolling features
        X_new: np.ndarray = self._transformer.transform(X)

        # add original data if wanted
        if self.passthrough:
            X_new = np.hstack([X.values, X_new])

        # put in dataframe
        X_new = pd.DataFrame(X_new, columns=self.feature_names_full_, index=X.index)

        # delete time features here, as we only want to return the rollings
        X_new = X_new.drop(self.timecols, axis=1)

        return X_new

    def compute_diagnostics(
        self,
        X_train: pd.DataFrame,
        X_test: pd.DataFrame,
        y_train: pd.DataFrame,
        y_test: pd.DataFrame,
    ) -> Tuple[float, float, np.ndarray, np.ndarray]:
        """
        This function is meant to provide some insight in the performance gained by adding the
        rolling features.

        For this, it computes r-squared between y_test and predictions made for two different proxy
        models of type defined by self.estimator_type.
        The first is for the original features presented in X_train and X_test (r2_base),
        and the second is for these features including the rolling features (r2_rollings).
        It also returns the fitted predictions.

        Note!: all input must be linearly increasing in time and have a datetime index.

        Parameters
        ----------
        X_train: pandas dataframe
            with shape [n_samples x n_features].
        X_test: pandas dataframe
            with shape [n_samples x n_features].
        y_train: pandas dataframe
            with shape [n_samples x n_features].
        y_test: pandas dataframe
            with shape [n_samples]

        Returns
        -------
        r2_base: float
            r-squared for the base model (without rollings)
        r2_rollings: float
            r-squared for the model including rollings
        yhat_base: 1D array
            prediction for the base model (without rollings)
        yhat_roll: 1D array
            prediction for the model including rollings
        """

        if not has_strictly_increasing_index(X_train):
            raise ValueError(INPUT_VALIDATION_ERR_MESSAGE)
        if not has_strictly_increasing_index(X_test):
            raise ValueError(INPUT_VALIDATION_ERR_MESSAGE)
        if not has_strictly_increasing_index(y_train):
            raise ValueError(INPUT_VALIDATION_ERR_MESSAGE)
        if not has_strictly_increasing_index(y_test):
            raise ValueError(INPUT_VALIDATION_ERR_MESSAGE)

        check_is_fitted(self, "feature_names_")

        # setup base estimation
        X_train_base, _ = self._add_time_features(X_train)
        X_test_base, _ = self._add_time_features(X_test)

        base_model: RegressorMixin = self._setup_estimator()
        base_model = base_model.fit(X_train_base, y_train)
        yhat_base: np.ndarray = base_model.predict(X_test_base)

        # and now with rolling features
        X_train_rolling: pd.DataFrame = self.transform(X_train)
        X_test_rolling: pd.DataFrame = self.transform(X_test)
        X_train_rolling, _ = self._add_time_features(X_train_rolling)
        X_test_rolling, _ = self._add_time_features(X_test_rolling)

        imputer = SimpleImputer()  # required as nans are created in rolling features
        X_train_rolling = imputer.fit_transform(X_train_rolling)
        X_test_rolling: pd.DataFrame = imputer.transform(X_test_rolling)

        roll_model: RegressorMixin = self._setup_estimator()
        roll_model = roll_model.fit(X_train_rolling, y_train)
        yhat_roll: np.ndarray = roll_model.predict(X_test_rolling)

        # compute r-squareds
        r2_base: float = r2_score(y_test, yhat_base)
        r2_rollings: float = r2_score(y_test, yhat_roll)

        return r2_base, r2_rollings, yhat_base, yhat_roll



from .custom_callbacks import R2Evaluation  # noqa: F401
from .incident_recall import (
    incident_recall,
    make_incident_recall_scorer,
    precision_incident_recall_curve,
)
from .keras_metrics import (  # noqa: F401
    get_keras_forecasting_metrics,
    keras_joint_mae_tilted_loss,
    keras_joint_mse_tilted_loss,
    keras_rmse,
    keras_tilted_loss,
)
from .mase import mean_absolute_scaled_error
from .quantile_evaluation import compute_quantile_crossings, compute_quantile_ratios
from .r2_calculation import train_mean_r2, train_r2
from .tilted_loss_metrics import (
    joint_mae_tilted_loss,
    joint_mse_tilted_loss,
    tilted_loss,
)

__all__ = [
    "incident_recall",
    "make_incident_recall_scorer",
    "precision_incident_recall_curve",
    "mean_absolute_scaled_error",
    "joint_mae_tilted_loss",
    "joint_mse_tilted_loss",
    "tilted_loss",
    "keras_tilted_loss",
    "keras_rmse",
    "get_keras_forecasting_metrics",
    "train_mean_r2",
    "train_r2",
    "compute_quantile_ratios",
    "compute_quantile_crossings",
]



from typing import Callable, Tuple

import pandas as pd
from sam.preprocessing import sam_format_to_wide
from sklearn.experimental import enable_iterative_imputer  # noqa F401
from sklearn.impute import IterativeImputer
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import train_test_split


def preprocess_data_for_benchmarking(
    data: pd.DataFrame,
    column_filter: Callable,
    targetcol: str,
    test_size: float = 0.3,
    resample: bool = "auto",
    resample_freq: str = "auto",
    ffill_limit: int = 5,
):
    """
    Takes a dataframe in SAM format, and converts it to X_train/X_test/y_train/y_test, while
    taking some liberties when it comes to 'faithfulness' of the underlying data.

    This function should NEVER be used in an actual project, but only for benchmarking models.
    For example, you want to compare two feature engineering methods to each other on a dataset in
    SAM format, but don't feel like properly reshaping/imputing/resampling the dataset yourself.

    Parameters
    ----------
    data: pd.DataFrame
        data in SAM format (TIME, TYPE, ID, VALUE columns)
    column_filter: function
        Function that takes in a column name (string) and returns a boolean, if this column should
        be included in the result.
        The columns 'TIME' and `targetcol` will be included in the result regardless of
        this function.
    targetcol: str
        The column to use as the target.
    test_size: float, optional (default=0.3)
        The portion of the data to use as testdata. This is always the last portion. So by default,
        the first 70% is train, the last 30% is test
    resample: bool, optional (default='auto')
        Whether or not to resample the data. If resample == 'auto', resample if the data is not
        monospaced yet
    resample_freq: str, optional (default='auto')
        If resample, the frequency to resample to. If resample_freq == 'auto', use the median
        frequency of the data
    ffill_limit: int, optional (default=5)
        If resample, the maximum number of values to ffill. By default, only ffill 5 values.
        This means the data won't be exactly monospaced, but will prevent extremely long flatlines
        if there are gaps in the data

    Returns
    ----------
    X_train: pd.DataFrame
        Monospaced data in SAM format with no `targetcol` that can be used for training
        together with y_train
    X_test: pd.DataFrame
        Monospaced data in SAM format with no `targetcol` that can be used for testing
        together with y_test
    y_train: pd.Series
        Values from `targetcol` that can be used for training together with X_train
    y_test: pd.Series
        Values from `targetcol` that can be used for testing together with X_test
    """
    # Convert to wide format
    data = sam_format_to_wide(data)
    # Select only relevant columns
    data = data[filter(lambda x: x == "TIME" or x == targetcol or column_filter(x), data.columns)]
    cols = data.columns  # Backup for after imputer\

    if resample == "auto":
        monospaced = data["TIME"].diff()[1:].unique().size == 1
        resample = not monospaced
    if resample and resample_freq == "auto":
        resample_freq = data["TIME"].diff()[1:].median()

    if resample:
        data = (
            data.set_index("TIME")
            .resample(resample_freq)
            .pad(limit=ffill_limit)
            .dropna(how="all")
            .reset_index()
        )

    # Input values
    time = data.pop("TIME")
    data = IterativeImputer().fit_transform(data)
    # Recover the column structure and the TIME column
    data = pd.DataFrame(data, columns=[col for col in cols if col != "TIME"])
    data["TIME"] = time

    # Now create train/test split
    y = data.pop(targetcol)
    X_train, X_test, y_train, y_test = train_test_split(
        data, y, test_size=test_size, shuffle=False
    )

    return X_train, X_test, y_train, y_test


def benchmark_model(
    train_test_data: Tuple[pd.DataFrame, pd.DataFrame, pd.Series, pd.Series],
    scorer: Callable = mean_squared_error,
    validation_data=True,
    return_histories=False,
    fit_kwargs=None,
    **modeldict,
):
    """
    Benchmarks a dictionary of sam models on train/test data, and returns a dictionary with scores
    The models are assumed to be SAM models in 2 ways:

        - `predict` is called as `predict(X_test, y_test)`
        - `model.get_actual(y_test)` is called

    Parameters
    ----------
    train_test_data: tuple
        tuple with elements X_train, X_test, y_train, y_test
    scorer: function, optional (default=sklearn.metrics.mean_squared_error)
        Function with signature ``func(y_true, y_pred)``
        where ``y_true`` and ``y_pred`` are pandas series, and it returns a scalar
    validation_data: bool
        If true, X_test and y_test will be added as 'validation_data' in the kwargs
        (fit_kwargs in this case) when fitting the models.
    return_histories: bool (default=False)
        If true, returns also a dictionary of History objects
    fit_kwargs: dict (default=None)
        The kwargs to include in the method fit of each model.
    modeldict: dict
        Dictionary the model names as keys (str) and the model objects as values

    Returns
    -------
    dict:
        Dictionary of form: ``{modelname_1: score, ..., modelname_n: score,
        persistence_benchmark: score, mean_benchmark: score}``
        where all scores are scalars.
    dict (optional):
        If return_histories == True returns also a dictionary of tensorflow History objects
    """
    if fit_kwargs is None:
        fit_kwargs = {}

    X_train, X_test, y_train, y_test = train_test_data

    if validation_data:
        fit_kwargs["validation_data"] = (X_test, y_test)
    final_scores = {}
    histories = {}

    for modelname in modeldict:

        model = modeldict[modelname]

        history = model.fit(X_train, y_train, **fit_kwargs)
        pred = model.predict(X_test, y_test)

        results = pd.DataFrame({"pred": pred, "actual": model.get_actual(y_test)}).dropna(axis=0)

        score = scorer(results["actual"], results["pred"])
        final_scores[modelname] = score
        histories[modelname] = history

    # Benchmark results:
    # We use the last model to get actual
    results = pd.DataFrame(
        {
            "actual": model.get_actual(y_test),
            "persistence_benchmark": y_test,
            "mean_benchmark": y_test.mean(),
        }
    ).dropna(axis=0)

    final_scores["persistence_benchmark"] = scorer(
        results["actual"], results["persistence_benchmark"]
    )
    final_scores["mean_benchmark"] = scorer(results["actual"], results["mean_benchmark"])

    if return_histories:
        return final_scores, histories
    else:
        return final_scores


def plot_score_dicts(**score_dicts):
    """
    Very simple plotting function for showing the results

    Parameters
    ----------
    score_dicts: kwargs
       Containing score dictionaries (output from `benchmark_model`). The key of the dictionary
       is the name/descriptor of the dataset.

    Returns
    --------
    matplotlib.axes.Axes:
        Plot of score dictionaries

    Examples
    --------
    >>> fig = plot_score_dicts(
    ...     chicago={'model_a': 0.5, 'persistence_benchmark': 0.6, 'mean_benchmark': 0.4},
    ...     china={'model_a': 0.1, 'mean_benchmark': 0.3, 'persistence_benchmark': 0.4}
    ... )
    """
    return pd.DataFrame(score_dicts).transpose().plot(kind="bar")


def benchmark_wrapper(
    models: dict,
    datasets: dict,
    column_filters: dict,
    targetcols: dict,
):
    """
    Wrapper around entire benchmark pipeline.
    Takes a dictionary of models, dictionary of datasets in SAM format,
    and preprocesses, fits and evaluates all models and benchmarks, and plots them

    Parameters
    ----------
    models: dict
        Dictionary of SAM models
    datasets: dict
        Dictionary of datasets in SAM format
    column_filters: dict
        Dictionary of functions that accept a colum in ID_TYPE format
    targetcols: dict
        Dictionary of strings, the targetcolumns in ID_TYPE format

    Examples
    --------
    >>> from sam.datasets import load_rainbow_beach, load_sewage_data
    >>> from sam.models import MLPTimeseriesRegressor, benchmark_wrapper
    >>> from sam.preprocessing import wide_to_sam_format
    >>>
    >>> sewage = load_sewage_data()
    >>> sewage = sewage.drop(['Precipitation', 'Temperature'], axis=1)
    >>> sewage = sewage.iloc[0:200, :]
    >>> sewage['TIME'] = sewage.index
    >>> sewage = wide_to_sam_format(sewage)
    >>> datasets = {
    ...     'sewage': sewage,
    ... }
    >>> column_filters = {
    ...     'sewage': lambda x: x,
    ... }
    >>> targetcols = {
    ...     'sewage': 'Discharge_Hoofdgemaal',
    ... }
    >>> models = {
    ...     'mymodel': MLPTimeseriesRegressor(predict_ahead=[3], timecol='TIME',
    ...                               dropout=0.5, verbose=True)  # some non-default params
    ... }
    >>> benchmark_wrapper(models, datasets, column_filters, targetcols)  # doctest: +ELLIPSIS
    Epoch ...
    """
    data_names = datasets.keys()
    traintest_dict = dict(
        [
            (
                name,
                preprocess_data_for_benchmarking(
                    datasets[name], column_filters[name], targetcols[name]
                ),
            )
            for name in data_names
        ]
    )
    scores = {}
    for name in data_names:
        scores[name] = benchmark_model(traintest_dict[name], **models)
    plot_score_dicts(**scores)



from typing import Callable

import numpy as np
import pandas as pd

from sam.models.base_model import BaseTimeseriesRegressor


class SamShapExplainer(object):
    """
    An object that imitates a SHAP explainer object. (Sort of) implements the base Explainer
    interface which can be found here
    <https://github.com/slundberg/shap/blob/master/shap/explainers/explainer.py>.
    The more advanced, tensorflow-specific attributes can be accessed with obj.explainer.
    The reason the interface is only sort of implemented, is the same reason why
    MLPTimeseriesRegressor doesn't entirely implement the skearn interface - for predicting, y is
    needed, which is not supported by the SamShapExplainer.

    Parameters
    ----------
    explainer: shap TFDeepExplainer object
        A shap explainer object. This will be used to generate the actual shap values
    model: BaseTimeseriesRegressor model
        This will be used to do the preprocessing before calling explainer.shap_values
    """

    def __init__(
        self, explainer: Callable, model: BaseTimeseriesRegressor, preprocess_predict: Callable
    ) -> None:
        self.explainer = explainer
        self.preprocess_predict = preprocess_predict

        # Create a proxy model that can call only 3 attributes we need
        class SamProxyModel:
            fit = None
            feature_names_ = model.get_feature_names_out()
            preprocess_predict = model.preprocess_predict

        self.model = SamProxyModel()
        # Trick sklearn into thinking this is a fitted variable
        self.model.feature_engineer_ = model.feature_engineer_
        # Will likely be somewhere around 0
        self.expected_value = explainer.expected_value

    def shap_values(self, X: pd.DataFrame, y: pd.Series = None, *args, **kwargs) -> np.ndarray:
        """
        Imitates explainer.shap_values, but combined with the preprocessing from the model.
        Returns a similar format as a regular shap explainer: a list of numpy arrays, one
        for each output of the model.

        Parameters
        ----------
        X: pd.DataFrame
            The dataframe used to 'train' the explainer
        y: pd.Series, optional (default=None)
            Target data used to 'train' the explainer.
        """
        X_transformed = self.model.preprocess_predict(X, y, dropna=False)
        return self.explainer.shap_values(X_transformed, *args, **kwargs)

    def attributions(self, X: pd.DataFrame, y: pd.Series = None, *args, **kwargs) -> np.ndarray:
        """
        Imitates explainer.attributions, which by default just mirrors shap_values

        Parameters
        ----------
        X: pd.DataFrame
            The dataframe used to 'train' the explainer
        y: pd.Series, optional (default=None)
            Target data used to 'train' the explainer.
        """
        return self.shap_values(X, y, *args, **kwargs)

    def test_values(self, X: pd.DataFrame, y: pd.Series = None) -> pd.DataFrame:
        """
        Only the preprocessing from the model, without the shap values.
        Returns a pandas dataframe with the actual values used for the explaining
        Can be used to better interpret the numpy array that is outputted by shap_values.
        For example, if shap_values outputs a 5x20 numpy array, that means you explained 5 objects
        with 20 features. This function will then return a 5x20 pandas dataframe.

        Parameters
        ----------
        X: pd.DataFrame
            The dataframe used to 'train' the explainer
        y: pd.Series, optional (default=None)
            Target data used to 'train' the explainer.
        """
        X_transformed = self.model.preprocess_predict(X, y, dropna=False)
        return pd.DataFrame(X_transformed, columns=self.model.feature_names_, index=X.index)
